#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from cycling.loader import get_interval, get_interval_cls
from task_trigger import get_message_offset


class MessageOutput(object):
    """
    A task message output.

    Used to generate an output string for a message trigger at a cycle point.

    TODO - these can be plain strings once the deprecated cycle point offset
    placeholders are removed from cylc (see GitHub #1761).

    """

    def __init__(self, msg, base_interval=None):
        self.msg = msg
        self.msg_offset = get_message_offset(msg, base_interval)

    def get_string(self, point):
        """Return the message string for this cycle point.

        Placeholders are replaced with the actual cycle point offset.

        """
        new_point = point
        if self.msg_offset:
            new_point = point + self.msg_offset
        return re.sub('\[.*\]', str(new_point), self.msg)

    def __eq__(self, other):
        return self.msg == other.msg and self.msg_offset == other.msg_offset



# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# (C) British Crown Copyright 2013-2014 Met Office.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

"""This provides utilites for extracting the local time zone."""

import time


def get_local_time_zone():
    """Return the current local UTC offset in hours and minutes."""
    utc_offset_seconds = -time.timezone
    if time.localtime().tm_isdst == 1 and time.daylight:
        utc_offset_seconds = -time.altzone
    utc_offset_minutes = (utc_offset_seconds // 60) % 60
    utc_offset_hours = utc_offset_seconds // 3600
    return utc_offset_hours, utc_offset_minutes


def get_local_time_zone_format(extended_mode=False, reduced_mode=False):
    """Return a string denoting the current local UTC offset."""
    utc_offset_hours, utc_offset_minutes = get_local_time_zone()
    if utc_offset_hours == 0 and utc_offset_minutes == 0:
        return "Z"
    reduced_timezone_template = "%s%02d"
    timezone_template = "%s%02d%02d"
    if extended_mode:
        timezone_template = "%s%02d:%02d"
    sign = "-" if (utc_offset_hours < 0 or utc_offset_minutes < 0) else "+"
    if reduced_mode and utc_offset_minutes == 0:
        return reduced_timezone_template % (sign, abs(utc_offset_hours))
    return timezone_template % (
        sign, abs(utc_offset_hours), abs(utc_offset_minutes))



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import Pyro.core
import hashlib
import os
import sys

from Pyro.protocol import DefaultConnValidator
import Pyro.constants
import Pyro.errors
import hmac

from cylc.cfgspec.globalcfg import GLOBAL_CFG
from cylc.network import NO_PASSPHRASE, PRIVILEGE_LEVELS
from cylc.config import SuiteConfig
from cylc.suite_host import get_hostname, is_remote_host
from cylc.suite_logging import LOG
from cylc.owner import USER


# Access for users without the suite passphrase: encrypting the "no passphrase"
# passphrase is unnecessary, but doing so allows common passphrase handling.

OK_HASHES = GLOBAL_CFG.get()['authentication']['hashes']
SCAN_HASH = GLOBAL_CFG.get()['authentication']['scan hash']
if SCAN_HASH not in OK_HASHES:
    OK_HASHES.append(SCAN_HASH)


CONNECT_DENIED_TMPL = "[client-connect] DENIED %s@%s:%s %s"
CONNECT_ALLOWED_TMPL = "[client-connect] %s@%s:%s privilege='%s' %s"


class ConnValidator(DefaultConnValidator):
    """Custom Pyro connection validator for user authentication."""

    HASHES = {}
    LENGTH_HASH_DIGESTS = {}
    NO_PASSPHRASE_HASHES = {}

    def set_pphrase(self, pphrase):
        """Store encrypted suite passphrase (called by the server)."""
        self.pphrase_hashes = {}
        for hash_name in OK_HASHES:
            hash_ = self._get_hash(hash_name)
            self.pphrase_hashes[hash_name] = hash_(pphrase).hexdigest()

    def set_default_hash(self, hash_name):
        """Configure a hash to use as the default."""
        self._default_hash_name = hash_name
        if None in self.HASHES:
            self.HASHES.pop(None)  # Pop default setting.

    def acceptIdentification(self, daemon, connection, token, challenge):
        """Authorize client login."""

        is_old_client = False
        # Processes the token returned by createAuthToken.
        try:
            user, host, uuid, prog_name, proc_passwd = token.split(':', 4)
        except ValueError:
            # Back compat for old suite client (passphrase only)
            # (Allows old scan to see new suites.)
            proc_passwd = token
            is_old_client = True
            user = "(user)"
            host = "(host)"
            uuid = "(uuid)"
            prog_name = "(OLD_CLIENT)"

        hash_name = self._get_hash_name_from_digest_length(proc_passwd)

        if hash_name not in OK_HASHES:
            return (0, Pyro.constants.DENIED_SECURITY)

        hash_ = self._get_hash(hash_name)

        # Access for users without the suite passphrase: encrypting the
        # no-passphrase is unnecessary, but doing so allows common handling.
        no_passphrase_hash = self._get_no_passphrase_hash(hash_name)

        # Check username and password, and set privilege level accordingly.
        # The auth token has a binary hash that needs conversion to ASCII.
        if self._compare_hmacs(
                hmac.new(challenge,
                         self.pphrase_hashes[hash_name].decode("hex"),
                         hash_).digest(),
                proc_passwd):
            # The client has the suite passphrase.
            # Access granted at highest privilege level.
            priv_level = PRIVILEGE_LEVELS[-1]
        elif not is_old_client and self._compare_hmacs(
                hmac.new(challenge,
                         no_passphrase_hash.decode("hex"),
                         hash_).digest(),
                proc_passwd):
            # The client does not have the suite passphrase.
            # Public access granted at level determined by global/suite config.
            config = SuiteConfig.get_inst()
            priv_level = config.cfg['cylc']['authentication']['public']
        else:
            # Access denied.
            if not is_old_client:
                # Avoid logging large numbers of denials from old scan clients
                # that try all passphrases available to them.
                LOG.warn(CONNECT_DENIED_TMPL % (user, host, prog_name, uuid))
            return (0, Pyro.constants.DENIED_SECURITY)

        # Store client details for use in the connection thread.
        connection.user = user
        connection.host = host
        connection.prog_name = prog_name
        connection.uuid = uuid
        connection.privilege_level = priv_level
        LOG.debug(CONNECT_ALLOWED_TMPL % (
                  user, host, prog_name, priv_level, uuid))
        return (1, 0)

    def createAuthToken(self, authid, challenge, peeraddr, URI, daemon):
        """Return a secure auth token based on the server challenge string.

        Argument authid is what's returned by mungeIdent().

        """
        hash_ = self._get_hash()
        return ":".join(
            list(authid[:4]) +
            [hmac.new(challenge, authid[4], hash_).digest()]
        )

    def mungeIdent(self, ident):
        """Receive (uuid, passphrase) from client. Encrypt the passphrase.

        Also pass client identification info to server for logging:
        (user, host, prog name).

        """
        hash_ = self._get_hash()
        uuid, passphrase = ident
        prog_name = os.path.basename(sys.argv[0])
        if passphrase is None:
            passphrase = NO_PASSPHRASE
        return (USER, get_hostname(), str(uuid), prog_name,
                hash_(passphrase).digest())

    def _compare_hmacs(self, hmac1, hmac2):
        """Compare hmacs as securely as possible."""
        try:
            return hmac.compare_hmacs(hmac1, hmac2)
        except AttributeError:
            # < Python 2.7.7.
            return (hmac1 == hmac2)

    def _get_default_hash_name(self):
        if hasattr(self, "_default_hash_name"):
            return self._default_hash_name
        return GLOBAL_CFG.get()['authentication']['hashes'][0]

    def _get_hash(self, hash_name=None):
        try:
            return self.HASHES[hash_name]
        except KeyError:
            pass

        hash_name_dest = hash_name
        if hash_name is None:
            hash_name = self._get_default_hash_name()

        self.HASHES[hash_name_dest] = getattr(hashlib, hash_name)
        return self.HASHES[hash_name_dest]

    def _get_hash_name_from_digest_length(self, digest):
        if len(digest) in self.LENGTH_HASH_DIGESTS:
            return self.LENGTH_HASH_DIGESTS[len(digest)]
        for hash_name in OK_HASHES:
            hash_ = self._get_hash(hash_name)
            len_hash = len(hash_("foo").digest())
            self.LENGTH_HASH_DIGESTS[len_hash] = hash_name
            if len_hash == len(digest):
                return hash_name

    def _get_no_passphrase_hash(self, hash_name=None):
        try:
            return self.NO_PASSPHRASE_HASHES[hash_name]
        except KeyError:
            hash_ = self._get_hash(hash_name)
            self.NO_PASSPHRASE_HASHES[hash_name] = (
                hash_(NO_PASSPHRASE).hexdigest())
        return self.NO_PASSPHRASE_HASHES[hash_name]



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Cylc suite graphing module. Modules relying on this should test for
ImportError due to pygraphviz/graphviz not being installed."""

import re
import pygraphviz
from cylc.task_id import TaskID
from cycling.loader import get_point, get_point_relative, get_interval
from graphnode import graphnode

# TODO: Do we still need autoURL below?


class CGraphPlain(pygraphviz.AGraph):
    """Directed Acyclic Graph class for cylc dependency graphs."""

    def __init__(self, title, suite_polling_tasks={}):
        self.title = title
        pygraphviz.AGraph.__init__(self, directed=True, strict=True)
        # graph attributes
        # - label (suite name)
        self.graph_attr['label'] = title
        self.suite_polling_tasks = suite_polling_tasks

    def node_attr_by_taskname(self, node_string):
        try:
            name, point_string = TaskID.split(node_string)
        except ValueError:
            # Special node?
            if node_string.startswith("__remove_"):
                return []
            raise
        if name in self.task_attr:
            return self.task_attr[name]
        else:
            return []

    def style_edge(self, left, right):
        pass

    def style_node(self, node_string, autoURL, base=False):
        node = self.get_node(node_string)
        try:
            name, point_string = TaskID.split(node_string)
        except ValueError:
            # Special node?
            if node_string.startswith("__remove_"):
                node.attr['style'] = 'dashed'
                node.attr['label'] = u'\u2702'
                return
            raise
        label = name
        if name in self.suite_polling_tasks:
            label += "\\n" + self.suite_polling_tasks[name][3]
        label += "\\n" + point_string
        node.attr['label'] = label
        if autoURL:
            if base:
                # TODO - This is only called from cylc_add_edge in this
                # base class ... should it also be called from add_node?
                node.attr['URL'] = 'base:' + node_string
            else:
                node.attr['URL'] = node_string

    def cylc_add_node(self, node_string, autoURL, **attr):
        pygraphviz.AGraph.add_node(self, node_string, **attr)
        self.style_node(node_string, autoURL)

    def cylc_add_edge(self, left, right, autoURL, **attr):
        if left is None and right is None:
            pass
        elif left is None:
            self.cylc_add_node(right, autoURL)
        elif right is None:
            self.cylc_add_node(left, autoURL)
        else:
            pygraphviz.AGraph.add_edge(self, left, right, **attr)
            self.style_node(left, autoURL, base=True)
            self.style_node(right, autoURL, base=True)
            self.style_edge(left, right)

    def cylc_remove_nodes_from(self, nodes):
        """Remove nodes, returning extra edge structure if possible.

        Each group of connected to-be-removed nodes is replaced by a
        single special node to preserve dependency info between the
        remaining nodes.

        """
        if not nodes:
            return
        existing_nodes = set(self.nodes())
        add_edges = set()
        remove_nodes = set(nodes)
        remove_node_groups = {}
        groups = {}
        group_new_nodes = {}
        edges = self.edges()
        incoming_remove_edges = []
        outgoing_remove_edges = []
        internal_remove_edges = []
        for l_node, r_node in edges:
            if l_node in remove_nodes:
                if r_node in remove_nodes:
                    # This is an edge between connected nuke nodes.
                    internal_remove_edges.append((l_node, r_node))
                else:
                    # This is an edge between nuke and normal nodes.
                    outgoing_remove_edges.append((l_node, r_node))
            elif r_node in remove_nodes:
                incoming_remove_edges.append((l_node, r_node))

        if not outgoing_remove_edges:
            # Preserving edges doesn't matter - ditch this whole set.
            self.remove_nodes_from(nodes)
            return

        # Loop through all connected nuke nodes and group them up.
        group_num = -1
        for l_node, r_node in sorted(internal_remove_edges):
            l_group = remove_node_groups.get(l_node)
            r_group = remove_node_groups.get(r_node)
            if l_group is None:
                if r_group is None:
                    # Create a new group for l_node and r_node.
                    group_num += 1
                    groups[group_num] = set((l_node, r_node))
                    remove_node_groups[l_node] = group_num
                    remove_node_groups[r_node] = group_num
                else:
                    # r_node already in a group, l_node not - add l_node.
                    groups[r_group].add(l_node)
                    remove_node_groups[l_node] = r_group
            elif r_group is None:
                # l_node already in a group, r_node not - add r_node.
                groups[l_group].add(r_node)
                remove_node_groups[r_node] = l_group
            elif l_group != r_group:
                # They are members of different groups - combine them.
                for node in groups[r_group]:
                    remove_node_groups[node] = l_group
                groups[l_group] = groups[l_group].union(groups[r_group])
                groups.pop(r_group)
        # Some nodes are their own group and don't have connections.
        for node in nodes:
            if node not in remove_node_groups:
                # The node is its own group.
                group_num += 1
                groups[group_num] = set([node])
                remove_node_groups[node] = group_num

        # Consolidate all groups with the same in/out edges.
        group_edges = {}
        for l_node, r_node in incoming_remove_edges:
            r_group = remove_node_groups[r_node]
            group_edges.setdefault(r_group, [set(), set()])
            group_edges[r_group][0].add(l_node)

        for l_node, r_node in outgoing_remove_edges:
            l_group = remove_node_groups[l_node]
            group_edges.setdefault(l_group, [set(), set()])
            group_edges[l_group][1].add(r_node)

        for group1 in sorted(group_edges):
            if group1 not in group_edges:
                continue
            for group2 in sorted(group_edges):
                if (group1 != group2 and
                        group_edges[group1][0] == group_edges[group2][0] and
                        group_edges[group1][1] == group_edges[group2][1]):
                    # Both groups have the same incoming and outgoing edges.
                    for node in groups[group2]:
                        remove_node_groups[node] = group1
                    groups[group1] = groups[group1].union(groups[group2])
                    groups.pop(group2)
                    group_edges.pop(group2)

        # Create a new node name for the group.
        names = set()
        index = -1
        for group, group_nodes in sorted(groups.items()):
            index += 1
            name = "__remove_%s__" % index
            while name in existing_nodes or name in names:
                index += 1
                name = "__remove_%s__" % index
            group_new_nodes[group] = name
            names.add(name)

        new_edges = set()
        groups_have_outgoing = set()
        for l_node, r_node in outgoing_remove_edges:
            new_l_group = remove_node_groups[l_node]
            new_l_node = group_new_nodes[new_l_group]
            new_edges.add((new_l_node, r_node, True, False, False))
            groups_have_outgoing.add(new_l_group)

        for l_node, r_node in incoming_remove_edges:
            new_r_group = remove_node_groups[r_node]
            new_r_node = group_new_nodes[new_r_group]
            if new_r_group not in groups_have_outgoing:
                # Skip any groups that don't have edges on to normal nodes.
                continue
            new_edges.add((l_node, new_r_node, True, False, False))

        self.remove_nodes_from(nodes)
        self.add_edges(list(new_edges))

    def add_edges(self, edges, ignore_suicide=False):
        edges.sort()  # TODO: does sorting help layout stability?
        for edge in edges:
            left, right, skipped, suicide, conditional = edge
            if suicide and ignore_suicide:
                continue
            attrs = {}
            if conditional:
                if suicide:
                    attrs['style'] = 'dashed'
                    attrs['arrowhead'] = 'odot'
                else:
                    attrs['style'] = 'solid'
                    attrs['arrowhead'] = 'onormal'
            else:
                if suicide:
                    attrs['style'] = 'dashed'
                    attrs['arrowhead'] = 'dot'
                else:
                    attrs['style'] = 'solid'
                    attrs['arrowhead'] = 'normal'
            if skipped:
                # override
                attrs['style'] = 'dotted'
                attrs['arrowhead'] = 'oinv'

            attrs['penwidth'] = 2

            self.cylc_add_edge(
                left, right, True, **attrs
            )

    def add_cycle_point_subgraphs(self, edges):
        """Draw nodes within cycle point groups (subgraphs)."""
        point_string_id_map = {}
        for edge_entry in edges:
            for id_ in edge_entry[:2]:
                if id_ is None:
                    continue
                try:
                    point_string = TaskID.split(id_)[1]
                except IndexError:
                    # Probably a special node - ignore it.
                    continue
                point_string_id_map.setdefault(point_string, [])
                point_string_id_map[point_string].append(id_)
        for point_string, ids in point_string_id_map.items():
            self.add_subgraph(
                nbunch=ids, name="cluster_" + point_string,
                label=point_string, fontsize=28, rank="max", style="dashed"
            )

    def add_subgraph(self, nbunch=None, name=None, **attr):
        """Return subgraph induced by nodes in nbunch.

        Overrides (but does the same thing as) pygraphviz's
        AGraph.add_subgraph method.

        """

        name = name.encode(self.encoding)

        handle = pygraphviz.graphviz.agsubg(
            self.handle, name, 1)

        subgraph = pygraphviz.AGraph(
            handle=handle, name=name,
            strict=self.strict, directed=self.directed,
            **attr
        )

        nodes = self.prepare_nbunch(nbunch)
        subgraph.add_nodes_from(nodes)

        return subgraph


class CGraph(CGraphPlain):
    """Directed Acyclic Graph class for cylc dependency graphs.
    This class automatically adds node and edge attributes
    according to the suite.rc file visualization config."""

    def __init__(self, title, suite_polling_tasks={}, vizconfig={}):

        # suite.rc visualization config section
        self.vizconfig = vizconfig
        CGraphPlain.__init__(self, title, suite_polling_tasks)

        # graph attributes
        # - default node attributes
        for item in vizconfig['default node attributes']:
            attr, value = re.split('\s*=\s*', item)
            self.node_attr[attr] = value
        # - default edge attributes
        for item in vizconfig['default edge attributes']:
            attr, value = re.split('\s*=\s*', item)
            self.edge_attr[attr] = value

        # non-default node attributes by task name
        # TODO - ERROR CHECKING FOR INVALID TASK NAME
        self.task_attr = {}

        for item in self.vizconfig['node attributes']:
            if item in self.vizconfig['node groups']:
                # item is a group of tasks
                for task in self.vizconfig['node groups'][item]:
                    # for each task in the group
                    for attr in self.vizconfig['node attributes'][item]:
                        if task not in self.task_attr:
                            self.task_attr[task] = []
                        self.task_attr[task].append(attr)
            else:
                # item must be a task name
                for attr in self.vizconfig['node attributes'][item]:
                    if item not in self.task_attr:
                        self.task_attr[item] = []
                    self.task_attr[item].append(attr)

    def style_node(self, node_string, autoURL, base=False):
        super(self.__class__, self).style_node(
            node_string, autoURL, False)
        node = self.get_node(node_string)
        for item in self.node_attr_by_taskname(node_string):
            attr, value = re.split('\s*=\s*', item)
            node.attr[attr] = value
        if self.vizconfig['use node color for labels']:
            node.attr['fontcolor'] = node.attr['color']
        node.attr['penwidth'] = self.vizconfig['node penwidth']

    def style_edge(self, left, right):
        super(self.__class__, self).style_edge(left, right)
        left_node = self.get_node(left)
        edge = self.get_edge(left, right)
        if self.vizconfig['use node color for edges']:
            edge.attr['color'] = left_node.attr['color']
        elif self.vizconfig['use node fillcolor for edges']:
            if left_node.attr['style'] == 'filled':
                edge.attr['color'] = left_node.attr['fillcolor']
        edge.attr['penwidth'] = self.vizconfig['edge penwidth']


class edge(object):
    def __init__(self, left, right, sequence, suicide=False,
                 conditional=False):
        """contains qualified node names, e.g. 'foo[T-6]:out1'"""
        self.left = left
        self.right = right
        self.suicide = suicide
        self.sequence = sequence
        self.conditional = conditional

    def get_right(self, inpoint, start_point):
        inpoint_string = str(inpoint)
        if self.right is None:
            return None

        # strip off special outputs
        self.right = re.sub(':\w+', '', self.right)

        return TaskID.get(self.right, inpoint_string)

    def get_left(self, inpoint, start_point, base_interval):
        # strip off special outputs
        left = re.sub(':[\w-]+', '', self.left)

        left_graphnode = graphnode(left, base_interval=base_interval)
        if left_graphnode.offset_is_from_ict:
            point = get_point_relative(left_graphnode.offset_string,
                                       start_point)
        elif left_graphnode.offset_string:
            point = get_point_relative(left_graphnode.offset_string, inpoint)
        else:
            point = inpoint
        name = left_graphnode.name

        return TaskID.get(name, point)



#!usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import sys
import unittest
from copy import copy
from cylc.param_expand import GraphExpander

"""Module for parsing cylc graph strings."""


ARROW = '=>'


class GraphParseError(Exception):
    """For graph string parsing errors."""
    def __str__(self):
        # Restore some spaces for readability.
        return self.args[0].replace(ARROW, ' %s ' % ARROW)


class Replacement(object):
    """A class to remember match group information in re.sub() calls"""
    def __init__(self, replacement):
        self.replacement = replacement
        self.substitutions = []
        self.match_groups = []

    def __call__(self, match):
        matched = match.group(0)
        replaced = match.expand(self.replacement)
        self.substitutions.append((matched, replaced))
        self.match_groups.append(match.groups())
        return replaced


class GraphParser(object):
    """Class for extracting dependency information from cylc graph strings.

    For each task in the graph string, results are stored as:
        self.triggers[task_name][expression] = ([expr_task_names], suicide)
        self.original[task_name][expression] = original_expression

    (original_expression is separated out to allow comparison of triggers
    from different equivalent expressions, e.g. family vs member).

    This is currently intended to process a single multi-line graph string
    (i.e. the content of a single graph section). But it could be extended to
    store dependencies for the whole suite (call parse_graph multiple times
    and key results by graph section).

    The general form of a dependency is "EXPRESSION => NODE", where:
        * On the right, NODE is a task or family name
        * On the left, an EXPRESSION of nodes involving parentheses, and
          logical operators '&' (AND), and '|' (OR).
        * Node names may be parameterized (any number of parameters):
            NODE<i,j,k>
            NODE<i=0,j,k>  # specific parameter value
            NODE<i-1,j,k>  # offset parameter value
        * A parameterized qualified node name looks like this:
            NODE(<PARAMS>)([CYCLE-POINT-OFFSET])(:TRIGGER-TYPE)
        * The default trigger type is ':succeed'.
        * Trigger qualifiers are ignored on the right to allow chaining:
               "foo => bar => baz & qux"
          Think of this as describing the graph structure first, then
          annotating each node with a trigger type that is only meaningful on
          the left side of each pair (in the default ':succeed' case the
          trigger type can be ommitted, but it is still there in principle).
    """

    OP_AND = '&'
    OP_OR = '|'
    OP_AND_ERR = '&&'
    OP_OR_ERR = '||'
    SUICIDE_MARK = '!'
    TRIG_SUCCEED = ':succeed'
    TRIG_FAIL = ':fail'
    TRIG_FINISH = ':finish'
    FAM_TRIG_EXT_ALL = '-all'
    FAM_TRIG_EXT_ANY = '-any'
    LEN_FAM_TRIG_EXT_ALL = len(FAM_TRIG_EXT_ALL)
    LEN_FAM_TRIG_EXT_ANY = len(FAM_TRIG_EXT_ANY)

    # Match fully qualified parameterized single nodes.
    REC_NODE_FULL = re.compile(r'''
        (
        (?:(?:!)?\w[\w\-+%@]*)  # node name
        (?:<[\w,=\-+]+>)?       # optional parameter list
        (?:\[[\w\-\+\^]+\])?    # optional cycle point offset
        (?::[\w\-]+)?           # optional trigger type
        )
    ''', re.X)           # end of string

    # Extract node info from a left-side expression, after parameter expansion.
    REC_NODES = re.compile(r'''
        ((?:!)?\w[\w\-+%@]*)  # node name
        (\[[\w\-\+\^]+\])?    # optional cycle point offset
        (:[\w\-]+)?           # optional trigger type
    ''', re.X)

    REC_TRIG_QUAL = re.compile(r'''
        (?:(?:!)?\w[\w\-+%@]*)   # node name (ignore)
        (:[\w\-]+)?              # optional trigger type
    ''', re.X)

    REC_COMMENT = re.compile('#.*$')

    # Detect presence of expansion parameters in a graph line.
    REC_PARAMETERS = re.compile(r'<[\w,=\-+]+>')

    # Detect and extract suite state polling task info.
    REC_SUITE_STATE = re.compile('(\w+)(<([\w\.\-]+)::(\w+)(:\w+)?>)')

    def __init__(self, family_map=None, parameters=None, startup_tasks=[]):
        """Initializing the graph string parser.

        family_map (empty or None if no families) is:
            {family_name: [task member names]}
        parameters (empty or None if no parameters) is just passed on to the
        parameter expander classes (documented there).
        """
        self.family_map = family_map or {}
        self.parameters = parameters
        self.triggers = {}
        self.original = {}
        self.suite_state_polling_tasks = {}
        # Back-compat cylc-5:
        self.startup_tasks = startup_tasks
        self.get_startup = False
        self.r1_point = None
        self.startup_graph_text = ''

    def parse_graph(self, graph_string, get_startup=False, r1_point=None):
        """Parse the graph string for a single graph section.

        (Assumes any general line-continuation markers have been processed).
           1. Strip comments, whitespace, and blank lines.
              (all whitespace is removed up front so we don't have to consider
              it in regexes and strip it from matched elements)
           2. Join incomplete lines starting or ending with '=>'.
           3. Replicate and expand any parameterized lines.
           4. Split and process by pairs "left-expression => right-node":
              i. Replace families with members (any or all semantics).
             ii. Record parsed dependency information for each right-side node.
        """
        self.get_startup = get_startup
        self.r1_point = r1_point

        # Strip comments, whitespace, and blank lines.
        non_blank_lines = []
        for line in graph_string.split('\n'):
            line = self.__class__.REC_COMMENT.sub('', line)
            # Apparently this is the fastest way to strip all whitespace!:
            line = "".join(line.split())
            if not line:
                continue
            non_blank_lines.append(line)

        # Join incomplete lines (beginning or ending with an arrow).
        full_lines = []
        part_lines = []
        for i in range(0, len(non_blank_lines)):
            this_line = non_blank_lines[i]
            if i == 0:
                # First line can't start with an arrow.
                if this_line.startswith(ARROW):
                    raise GraphParseError(
                        "ERROR, leading arrow: %s" % this_line)
            try:
                next_line = non_blank_lines[i + 1]
            except IndexError:
                next_line = ''
                if this_line.endswith(ARROW):
                    # Last line can't end with an arrow.
                    raise GraphParseError(
                        "ERROR, trailing arrow: %s" % this_line)
            part_lines.append(this_line)
            if (this_line.endswith(ARROW) or next_line.startswith(ARROW)):
                continue
            full_line = ''.join(part_lines)

            # Record inter-suite dependence and remove the marker notation.
            # ("foo<SUITE::TASK:fail> => bar" becomes:fail "foo => bar").
            repl = Replacement('\\1')
            full_line = self.__class__.REC_SUITE_STATE.sub(repl, full_line)
            for item in repl.match_groups:
                l_task, r_all, r_suite, r_task, r_status = item
                if r_status:
                    r_status = r_status[1:]
                else:
                    r_status = self.__class__.TRIG_SUCCEED[1:]
                self.suite_state_polling_tasks[l_task] = (
                    r_suite, r_task, r_status, r_all)
            full_lines.append(full_line)
            part_lines = []

        # Check for double-char conditional operators (a common mistake),
        # and bad node syntax (order of qualifiers).
        bad_lines = []
        for line in full_lines:
            if self.__class__.OP_AND_ERR in line:
                raise GraphParseError(
                    "ERROR, the graph AND operator is '%s': %s" % (
                        self.__class__.OP_AND, line))
            if self.__class__.OP_OR_ERR in line:
                raise GraphParseError(
                    "ERROR, the graph OR operator is '%s': %s" % (
                        self.__class__.OP_OR, line))
            # Check node syntax. First drop all non-node characters.
            node_str = line
            for s in ['=>', '|', '&', '(', ')', '!']:
                node_str = node_str.replace(s, ' ')
            # Then drop all valid nodes, longest first to avoid sub-strings.
            nodes = self.__class__.REC_NODE_FULL.findall(node_str)
            nodes.sort(key=len, reverse=True)
            for node in nodes:
                node_str = node_str.replace(node, '')
            # Result should be empty string.
            if node_str.strip():
                bad_lines.append(line.strip())
        if bad_lines:
            raise GraphParseError(
                "ERROR, bad graph node format:\n"
                "  " + "\n  ".join(bad_lines) + "\n"
                "Correct format is"
                " NAME(<PARAMS>)([CYCLE-POINT-OFFSET])(:TRIGGER-TYPE)")

        # Expand parameterized lines (or detect undefined parameters).
        line_set = set()
        graph_expander = GraphExpander(self.parameters)
        for line in full_lines:
            if not self.__class__.REC_PARAMETERS.search(line):
                line_set.add(line)
                continue
            for l in graph_expander.expand(line):
                line_set.add(l)

        # Process chains of dependencies as pairs: left => right.
        # Parameterization can duplicate some dependencies, so use a set.
        pairs = set()
        for line in line_set:
            # "foo => bar => baz" becomes [foo, bar, baz]
            chain = line.split(ARROW)
            # Auto-trigger lone nodes and initial nodes in a chain.
            for name, offset, _ in self.__class__.REC_NODES.findall(chain[0]):
                if not offset:
                    pairs.add(('', name))
            for i in range(0, len(chain) - 1):
                pairs.add((chain[i], chain[i + 1]))

        for pair in pairs:
            self._proc_dep_pair(pair[0], pair[1])

        # If debugging, print the final result here:
        # self.print_triggers()

    def _proc_dep_pair(self, left, right):
        """Process a single dependency pair 'left => right'.

        'left' can be a logical expression of qualified node names.
        'right' can be one or more node names joined by AND.
        A node name is a task name or a family name.
        A qualified name is NAME([CYCLE-POINT-OFFSET])(:TRIGGER-TYPE).
        Trigger qualifiers, but not cycle offsets, are ignored on the right to
        allow chaining.
        """
        if self.__class__.OP_OR in right:
            raise GraphParseError("ERROR, illegal OR on RHS: %s" % right)
        # Remove aualifiers from right-side nodes.
        for qual in self.__class__.REC_TRIG_QUAL.findall(right):
            right = right.replace(qual, '')
        if self.__class__.SUICIDE_MARK in left:
            raise GraphParseError(
                "ERROR, suicide markers must be"
                " on the right of a trigger: %s" % left)
        # Cycle point offsets are not allowed on the right side (yet).
        if '[' in right:
            raise GraphParseError(
                "ERROR, illegal cycle point offset on the right: %s => %s" % (
                    left, right))
        # Check that parentheses match.
        if left.count("(") != left.count(")"):
            raise GraphParseError(
                "ERROR, parenthesis mismatch in: \"" + left + "\"")

        # Split right side on AND.
        rights = right.split(self.__class__.OP_AND)

        if self.__class__.OP_OR in left or '(' in left:
            # Treat conditional or bracketed expressions as a single entity.
            lefts = [left]
        else:
            # Split non-conditional left-side expressions on AND.
            lefts = left.split(self.__class__.OP_AND)

        for left in lefts:
            # Extract infomation about all nodes on the left.
            info = self.__class__.REC_NODES.findall(left)

            # Make success triggers explicit.
            n_info = []
            expr = left
            for name, offset, trig in info:
                if not trig:
                    trig = self.__class__.TRIG_SUCCEED
                    if offset:
                        this = r'\b%s\b%s(?!:)' % (name, re.escape(offset))
                    else:
                        this = r'\b%s\b(?![\[:])' % name

                    that = name + offset + trig
                    expr = re.sub(this, that, expr)
                n_info.append((name, offset, trig))
            info = n_info

            # Determine semantics of all family triggers present.
            families = {}
            for name, offset, trig in info:
                if name in self.family_map:
                    if trig.endswith(self.__class__.FAM_TRIG_EXT_ANY):
                        ttype = trig[:-self.__class__.LEN_FAM_TRIG_EXT_ANY]
                        ext = self.__class__.FAM_TRIG_EXT_ANY
                    elif trig.endswith(self.__class__.FAM_TRIG_EXT_ALL):
                        ttype = trig[:-self.__class__.LEN_FAM_TRIG_EXT_ALL]
                        ext = self.__class__.FAM_TRIG_EXT_ALL
                    else:
                        # Unqualified (FAM => foo) or bad (FAM:bad => foo).
                        raise GraphParseError(
                            "ERROR, bad family trigger in %s" % expr)
                    families[name] = (ttype, ext)
                else:
                    if (trig.endswith(self.__class__.FAM_TRIG_EXT_ANY) or
                            trig.endswith(self.__class__.FAM_TRIG_EXT_ALL)):
                        raise GraphParseError("ERROR, family trigger on non-"
                                              "family namespace %s" % expr)
            self._families_all_to_all(expr, rights, info, families)

    def _families_all_to_all(self, expr, rights, info, families):
        """Replace all family names with member names, for all/any semantics.

        (Also for graph segments with no family names.)
        """
        n_info = []
        n_expr = expr
        for name, offset, trig in info:
            if name in families:
                ttype, extn = families[name]
                m_info = []
                m_expr = []
                for mem in self.family_map[name]:
                    m_info.append((mem, offset, ttype))
                    m_expr.append("%s%s%s" % (mem, offset, ttype))
                this = r'\b%s%s%s\b' % (name, re.escape(offset), trig)
                if extn == self.__class__.FAM_TRIG_EXT_ALL:
                    that = '(%s)' % '&'.join(m_expr)
                elif extn == self.__class__.FAM_TRIG_EXT_ANY:
                    that = '(%s)' % '|'.join(m_expr)
                n_expr = re.sub(this, that, n_expr)
                n_info += m_info
            else:
                n_info += [(name, offset, trig)]
        info = n_info

        for right in rights:
            if right.startswith(self.__class__.SUICIDE_MARK):
                right = right[1:]
                suicide = True
            else:
                suicide = False
            if right in self.family_map:
                for mem in self.family_map[right]:
                    self._add_trigger(expr, mem, n_expr, info, suicide)
            else:
                self._add_trigger(expr, right, n_expr, info, suicide)

        self._tidy_triggers()

    def _add_trigger(self, orig_expr, right, expr, info, suicide):
        """Store trigger info from "expr => right" for a single task ('right').

        Arg info is [(name, offset, trigger_type)] for each name in expr.
        """
        if self.startup_tasks:
            self._add_trigger_back_compat(
                orig_expr, right, expr, info, suicide)
        else:
            self._add_trigger_new(orig_expr, right, expr, info, suicide)

    def _add_trigger_new(self, orig_expr, right, expr, info, suicide):
        """New __add_trigger() for cylc-7+ or if no cylc-5 start-up tasks."""
        trigs = []
        for name, offset, trigger in info:
            # Replace finish triggers (must be done after member substn).
            if trigger == self.__class__.TRIG_FINISH:
                this = "%s%s%s" % (name, re.escape(offset), trigger)
                that = "(%s%s%s%s%s%s%s)" % (
                    name, offset, self.__class__.TRIG_SUCCEED,
                    self.__class__.OP_OR,
                    name, offset, self.__class__.TRIG_FAIL)
                expr = re.sub(this, that, expr)
                trigs += [
                    "%s%s%s" % (name, offset, self.__class__.TRIG_SUCCEED),
                    "%s%s%s" % (name, offset, self.__class__.TRIG_FAIL)]
            else:
                this = "%s%s%s" % (name, re.escape(offset), trigger)
                that = "%s%s%s" % (name, offset, trigger)
                expr = re.sub(this, that, expr)
                trigs += ["%s%s%s" % (name, offset, trigger)]
        if right not in self.triggers:
            self.triggers[right] = {}
            self.original[right] = {}
        self.triggers[right][expr] = (trigs, suicide)
        self.original[right][expr] = orig_expr

    def _add_trigger_back_compat(self, orig_expr, right, expr, info, suicide):
        """Temporary (pre cylc-7) __add_trigger() for cylc-5 start-up tasks.

        Extract all start-up triggers for use in new R1 sections.
        """
        if right in self.startup_tasks:
            if not self.get_startup or self.r1_point:
                return
            elif self.r1_point and suicide:
                return
        # Replace finish triggers here (must be done after member substn).
        trigs = []
        pruned = []
        startup_found = False
        for name, offset, trigger in info:
            if name in self.startup_tasks:
                startup_found = True
            if (self.get_startup and name in self.startup_tasks and
                    self.r1_point and not offset):
                that_offset = '[^]'
            else:
                that_offset = offset
            if ((not self.get_startup and name in self.startup_tasks) or
                    (self.get_startup and name not in self.startup_tasks)):
                pruned.append("%s%s%s" % (name, that_offset, trigger))
            elif trigger == self.__class__.TRIG_FINISH:
                this = "%s%s%s" % (name, re.escape(offset), trigger)
                that = "(%s%s%s%s%s%s%s)" % (
                    name, that_offset, self.__class__.TRIG_SUCCEED,
                    self.__class__.OP_OR,
                    name, that_offset, self.__class__.TRIG_FAIL)
                expr = re.sub(this, that, expr)
                trigs += [
                    "%s%s%s" % (name,
                                that_offset, self.__class__.TRIG_SUCCEED),
                    "%s%s%s" % (name, that_offset, self.__class__.TRIG_FAIL)]
            else:
                this = "%s%s%s" % (name, re.escape(offset), trigger)
                that = "%s%s%s" % (name, that_offset, trigger)
                expr = re.sub(this, that, expr)
                trigs += ["%s%s%s" % (name, that_offset, trigger)]
        if (self.get_startup and not startup_found and
                right not in self.startup_tasks):
            return
        for node in pruned:
            expr = re.sub(re.escape(node) + '&', '', expr)
            expr = re.sub('&' + re.escape(node), '', expr)
            # TODO - IS THIS NEEDED?:
            expr = re.sub(re.escape(node), '', expr)
        if self.get_startup:
            if expr and right:
                self.startup_graph_text += '\n%s => %s' % (expr, right)
            elif right:
                self.startup_graph_text += '\n%s' % right
        if right not in self.triggers:
            self.triggers[right] = {}
            self.original[right] = {}
        self.triggers[right][expr] = (trigs, suicide)
        self.original[right][expr] = orig_expr

    def _tidy_triggers(self):
        """Remove any null triggers for tasks that also have non-null triggers.

        (Null triggers are only needed for initial or lone nodes.)
        """
        # NOT USED, but leaving this here for reference.
        # Extra null triggers have no adverse effect, but this method has a big
        # negative performance impact on the dev/suites/busy.
        return

        clean_me = set()
        for right, val in self.triggers.items():
            exprs = val.keys()
            if len(exprs) > 1:
                if '' in exprs:
                    clean_me.add(right)
        for right in clean_me:
            del self.triggers[right]['']

    def print_triggers(self):
        for right, val in self.triggers.items():
            for expr, info in val.items():
                triggers, suicide = info
                if suicide:
                    title = "SUICIDE:"
                else:
                    title = "TRIGGER:"
                print '\nTASK:', right
                print ' ', title, expr
                for t in triggers:
                    print '    +', t
                print '  from', self.original[right][expr]


class TestGraphParser(unittest.TestCase):
    """Unit tests for the GraphParser class."""

    def test_line_continuation(self):
        """Test syntax-driven line continuation."""
        graph1 = "a => b => c"
        graph2 = """a =>
 b => c"""
        graph3 = """a => b
 => c"""
        gp1 = GraphParser()
        gp1.parse_graph(graph1)
        gp2 = GraphParser()
        gp2.parse_graph(graph2)
        gp3 = GraphParser()
        gp3.parse_graph(graph3)
        res = {
            'a': {'': ([], False)},
            'c': {'b:succeed': (['b:succeed'], False)},
            'b': {'a:succeed': (['a:succeed'], False)}
        }
        self.assertEqual(gp1.triggers, res)
        self.assertEqual(gp1.triggers, gp2.triggers)
        self.assertEqual(gp1.triggers, gp3.triggers)
        graph = """foo => bar
            a => b =>"""
        gp = GraphParser()
        self.assertRaises(GraphParseError, gp.parse_graph, graph)
        graph = """ => a => b
            foo => bar"""
        gp = GraphParser()
        self.assertRaises(GraphParseError, gp.parse_graph, graph)

    def test_def_trigger(self):
        """Test default trigger is :succeed."""
        gp1 = GraphParser()
        gp1.parse_graph("foo => bar")
        gp2 = GraphParser()
        gp2.parse_graph("foo:succeed => bar")
        self.assertEqual(gp1.triggers, gp2.triggers)

    def test_finish_trigger(self):
        """Test finish trigger expansion."""
        gp1 = GraphParser()
        gp1.parse_graph("foo:finish => bar")
        gp2 = GraphParser()
        gp2.parse_graph("(foo:succeed | foo:fail) => bar")
        self.assertEqual(gp1.triggers, gp2.triggers)

    def test_fam_all_to_all(self):
        """Test family all-to-all semantics."""
        fam_map = {'FAM': ['m1', 'm2'], 'BAM': ['b1', 'b2']}
        gp1 = GraphParser(fam_map)
        gp1.parse_graph("FAM:succeed-all => BAM")
        gp2 = GraphParser(fam_map)
        gp2.parse_graph("""
            (m1 & m2) => b1
            (m1 & m2) => b2""")
        self.assertEqual(gp1.triggers, gp2.triggers)

    def test_fam_one_to_all(self):
        """Test family one-to-all semantics."""
        fam_map = {'FAM': ['m1', 'm2']}
        gp1 = GraphParser(fam_map)
        gp1.parse_graph("pre => FAM")
        gp2 = GraphParser(fam_map)
        gp2.parse_graph("""
            pre => m1
            pre => m2
            """)
        self.assertEqual(gp1.triggers, gp2.triggers)

    def test_fam_all_to_one(self):
        """Test family all-to-one semantics."""
        fam_map = {'FAM': ['m1', 'm2']}
        gp1 = GraphParser(fam_map)
        gp1.parse_graph("FAM:succeed-all => post")
        gp2 = GraphParser(fam_map)
        gp2.parse_graph("(m1 & m2) => post")
        self.assertEqual(gp1.triggers, gp2.triggers)

    def test_fam_any_to_one(self):
        """Test family any-to-one semantics."""
        fam_map = {'FAM': ['m1', 'm2']}
        gp1 = GraphParser(fam_map)
        gp1.parse_graph("FAM:succeed-any => post")
        gp2 = GraphParser(fam_map)
        gp2.parse_graph("(m1 | m2) => post")
        self.assertEqual(gp1.triggers, gp2.triggers)

    def test_fam_any_to_all(self):
        """Test family any-to-all semantics."""
        fam_map = {'FAM': ['m1', 'm2'], 'BAM': ['b1', 'b2']}
        gp1 = GraphParser(fam_map)
        gp1.parse_graph("FAM:fail-any => BAM")

        gp2 = GraphParser(fam_map)
        gp2.parse_graph("""
            (m1:fail | m2:fail) => b1
            (m1:fail | m2:fail) => b2""")
        self.assertEqual(gp1.triggers, gp2.triggers)

    def test_fam_finish(self):
        """Test family finish semantics."""
        fam_map = {'FAM': ['m1', 'm2']}
        gp1 = GraphParser(fam_map)
        gp1.parse_graph("FAM:finish-all => post")
        gp2 = GraphParser(fam_map)
        gp2.parse_graph("""
            ((m1:succeed | m1:fail) & (m2:succeed | m2:fail)) => post""")
        self.assertEqual(gp1.triggers, gp2.triggers)

    def test_parameter_expand(self):
        """Test graph parameter expansion."""
        fam_map = {
            'FAM_m0': ['fa_m0', 'fb_m0'],
            'FAM_m1': ['fa_m1', 'fb_m1'],
        }
        params = {'m': ['0', '1'], 'n': ['0', '1']}
        templates = {'m': '_m%(m)s', 'n': '_n%(n)s'}
        gp1 = GraphParser(fam_map, (params, templates))
        gp1.parse_graph("""
            pre => foo<m,n> => bar<n>
            bar<n=0> => baz  # specific case
            bar<n-1> => bar<n>  # inter-chunk
            """)
        gp2 = GraphParser()
        gp2.parse_graph("""
            pre => foo_m0_n0 => bar_n0
            pre => foo_m0_n1 => bar_n1
            pre => foo_m1_n0 => bar_n0
            pre => foo_m1_n1 => bar_n1
            bar_n0 => baz
            bar_n0 => bar_n1
            """)
        self.assertEqual(gp1.triggers, gp2.triggers)

    def test_parameter_specific(self):
        """Test graph parameter expansion with a specific value."""
        params = {'i': ['0', '1'], 'j': ['0', '1', '2']}
        templates = {'i': '_i%(i)s', 'j': '_j%(j)s'}
        gp1 = GraphParser(family_map=None, parameters=(params, templates))
        gp1.parse_graph("bar<i-1,j> => baz<i,j>\nfoo<i=1,j> => qux")
        gp2 = GraphParser()
        gp2.parse_graph("""
           baz_i0_j0
           baz_i0_j1
           baz_i0_j2
           foo_i1_j0 => qux
           foo_i1_j1 => qux
           foo_i1_j2 => qux
           bar_i0_j0 => baz_i1_j0
           bar_i0_j1 => baz_i1_j1
           bar_i0_j2 => baz_i1_j2""")
        self.assertEqual(gp1.triggers, gp2.triggers)

    def test_parameter_offset(self):
        """Test graph parameter expansion with an offset."""
        params = {'i': ['0', '1'], 'j': ['0', '1', '2']}
        templates = {'i': '_i%(i)s', 'j': '_j%(j)s'}
        gp1 = GraphParser(family_map=None, parameters=(params, templates))
        gp1.parse_graph("bar<i-1,j> => baz<i,j>")
        gp2 = GraphParser()
        gp2.parse_graph("""
           baz_i0_j0
           baz_i0_j1
           baz_i0_j2
           bar_i0_j0 => baz_i1_j0
           bar_i0_j1 => baz_i1_j1
           bar_i0_j2 => baz_i1_j2""")
        self.assertEqual(gp1.triggers, gp2.triggers)

    def test_conditional(self):
        """Test generation of conditional triggers."""
        gp1 = GraphParser()
        gp1.parse_graph("(foo:start | bar) => baz")
        res = {
            'baz': {
                '(foo:start|bar:succeed)': (
                    ['foo:start', 'bar:succeed'], False)
            },
            'foo': {
                '': ([], False)
            },
            'bar': {
                '': ([], False)
            }
        }
        self.assertEqual(gp1.triggers, res)

    def test_repeat_trigger(self):
        """Test that repeating a trigger has no effect."""
        gp1 = GraphParser()
        gp2 = GraphParser()
        gp1.parse_graph("foo => bar")
        gp2.parse_graph("""
            foo => bar
            foo => bar""")
        self.assertEqual(gp1.triggers, gp2.triggers)

    def test_chain_stripping(self):
        """Test that trigger type is stripped off right-side nodes."""
        gp1 = GraphParser()
        gp1.parse_graph("""
        bar
        foo => bar:succeed => baz""")
        gp2 = GraphParser()
        gp2.parse_graph("""
            foo => bar
            bar:succeed => baz""")
        self.assertEqual(gp1.triggers, gp2.triggers)

    def test_double_oper(self):
        """Test that illegal forms of the logical operators are detected."""
        graph = "foo && bar => baz"
        gp = GraphParser()
        self.assertRaises(GraphParseError, gp.parse_graph, graph)
        graph = "foo || bar => baz"
        gp = GraphParser()
        self.assertRaises(GraphParseError, gp.parse_graph, graph)

    def test_bad_node_syntax(self):
        """Test that badly formatted graph nodes are detected.

        The correct format is:
          NAME(<PARAMS>)([CYCLE-POINT-OFFSET])(:TRIGGER-TYPE)")
        """
        graphs = [
            "foo[-P1Y]<m,n> => bar",
            "foo[-P1Y]<m,n> => bar",
            "foo:fail[-P1Y] => bar",
            "foo[-P1Y]:fail<m,n> => bar",
            "foo[-P1Y]<m,n>:fail => bar",
            "foo<m,n>:fail[-P1Y] => bar",
            "foo:fail<m,n>[-P1Y] => bar"
        ]
        detected = 0
        for graph in graphs:
            gp = GraphParser()
            try:
                gp.parse_graph(graph)
            except GraphParseError:
                detected += 1
        self.assertEqual(detected, len(graphs))

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Redirect stdout and stderr to the suite logs."""

import os
import sys

from cylc.cfgspec.globalcfg import GLOBAL_CFG
from cylc.suite_logging import OUT


class SuiteOutput(object):
    """Redirects stdout and stderr to the out, err files in the suite logging
    directory."""

    def __init__(self, suite):
        sodir = GLOBAL_CFG.get_derived_host_item(suite, 'suite log directory')
        self.opath = os.path.join(sodir, 'out')
        self.epath = os.path.join(sodir, 'err')

    def get_path(self, err=False):
        """Returns the path to the suite 'out' file or 'err' if err=True."""
        if err:
            return self.epath
        else:
            return self.opath

    def redirect(self):
        """redirect the standard file descriptors to suite log files."""
        # record current standard file descriptors
        self.sys_stdout = sys.stdout
        self.sys_stderr = sys.stderr
        self.sys_stdin = sys.stdin

        # redirect standard file descriptors
        # note that simply reassigning the sys streams is not sufficient
        # if we import modules that write to stdin and stdout from C
        # code - evidently the subprocess module is in this category!
        sout = file(self.opath, 'a+', 0)  # 0 => unbuffered
        serr = file(self.epath, 'a+', 0)
        dvnl = file('/dev/null', 'r')
        os.dup2(sout.fileno(), sys.stdout.fileno())
        os.dup2(serr.fileno(), sys.stderr.fileno())
        os.dup2(dvnl.fileno(), sys.stdin.fileno())

    def restore(self):
        """Restores stdout, err to their normal output."""
        # (not used)
        sys.stdout.close()
        sys.stderr.close()
        sys.stdout = self.sys_stdout
        sys.stderr = self.sys_stderr
        sys.stdin = self.sys_stdin
        OUT.info("Restored stdout and stderr to normal")



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time


def tail(file):
    while True:
        where = file.tell()
        line = file.readline()
        if not line:
            time.sleep(1)
            file.seek(where)
            # return even if no new line so the host thread doesn't hang when
            # the gui exits.
            yield None
        else:
            yield line



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gtk

from cylc.gui.dot_maker import DotMaker
from cylc.gui.util import get_icon
from cylc.task_state import TASK_STATUSES_ORDERED


class ThemeLegendWindow(gtk.Window):

    """This is a popup window displaying the theme state colors."""

    def __init__(self, parent_window, theme_map, dot_size='medium'):
        super(ThemeLegendWindow, self).__init__()
        self.set_border_width(5)
        self.set_title("")
        if parent_window is None:
            self.set_icon(get_icon())
        else:
            self.set_transient_for(parent_window)
        self.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_DIALOG)

        vbox = gtk.VBox()

        self._theme = theme_map
        self._dot_size = dot_size
        self._key_liststore = gtk.ListStore(str, gtk.gdk.Pixbuf)
        treeview = gtk.TreeView(self._key_liststore)
        treeview.set_headers_visible(False)
        treeview.get_selection().set_mode(gtk.SELECTION_NONE)
        tvc = gtk.TreeViewColumn(None)

        self.update()

        cellpb = gtk.CellRendererPixbuf()
        cell = gtk.CellRendererText()

        tvc.pack_start(cellpb, False)
        tvc.pack_start(cell, True)

        tvc.set_attributes(cellpb, pixbuf=1)
        tvc.set_attributes(cell, text=0)

        treeview.append_column(tvc)

        self.add(treeview)
        self.show_all()

    def update(self, new_theme=None, new_dot_size=None):
        """Update, optionally with a new theme."""
        if new_theme is not None:
            self._theme = new_theme
        if new_dot_size is not None:
            self._dot_size = new_dot_size
        self._set_key_liststore()

    def _set_key_liststore(self):
        dotm = DotMaker(self._theme, self._dot_size)
        self._key_liststore.clear()
        for state in TASK_STATUSES_ORDERED:
            dot = dotm.get_icon(state)
            self._key_liststore.append([state, dot])



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from cylc.suite_logging import OUT

# BROKER:
# A collection of output messages with associated owner ids (of the
# originating tasks) representing the outputs of ALL TASKS in the
# suite, and initialised from the outputs of all the tasks.
# "Satisfied" => the output has been completed.


class broker(object):
    # A broker aggregates output messages from many objects.
    # Each task registers its outputs with the suite broker, then each
    # task tries to get its prerequisites satisfied by the broker's
    # outputs.

    def __init__(self):
        self.all_outputs = {}   # all_outputs[ message ] = taskid
        self.all_output_msgs = set()

    def register(self, tasks):

        for task in tasks:
            self.all_outputs.update(task.state.outputs.completed)
            # TODO - SHOULD WE CHECK FOR SYSTEM-WIDE DUPLICATE OUTPUTS?
            # (note that successive tasks of the same type can register
            # identical outputs if they write staggered restart files).
        self.all_output_msgs = set(self.all_outputs)

    def reset(self):
        # throw away all messages
        self.all_outputs = {}
        self.all_output_msgs = set()

    def dump(self):
        # for debugging
        log_msg = "BROKER DUMP:"
        for msg in self.all_outputs:
            log_msg += "\n+ " + self.all_outputs[msg] + '\t' + msg
        OUT.info(log_msg)

    def negotiate(self, task):
        # can my outputs satisfy any of task's prerequisites
        task.state.satisfy_me(self.all_output_msgs, self.all_outputs)



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Process pool to execute shell commands for the suite daemon.

In debug mode, commands are printed to stdout before execution.

Some notes:
 * To print worker process ID, call multiprocessing.current_process()
 * To use a thread pool instead of a process pool:
   - use multiprocessing.pool.ThreadPool instead of pool.Pool
   - use multiprocessing.dummy.current_process() for current_process()
   - note that lsof logic may be required to ensure job scripts are
     closed properly after writing (see jobfile.py prior to cylc-6).
  (early versions of this module gave a choice of process or thread).
"""

import fileinput
import logging
import multiprocessing
from pipes import quote
from subprocess import Popen, PIPE
import sys
from tempfile import TemporaryFile
import time
import traceback

from cylc.batch_sys_manager import BATCH_SYS_MANAGER
from cylc.cfgspec.globalcfg import GLOBAL_CFG
import cylc.flags
from cylc.suite_logging import LOG, OUT
from cylc.wallclock import get_current_time_string


def _run_command(ctx):
    """Execute a shell command and capture its output and exit status."""

    if cylc.flags.debug:
        if ctx.cmd_kwargs.get('shell'):
            OUT.debug(ctx.cmd)
        else:
            OUT.debug(
                "%s\n" % ' '.join([quote(cmd_str) for cmd_str in ctx.cmd]))

    if (SuiteProcPool.STOP_JOB_SUBMISSION.value and
            ctx.cmd_key == SuiteProcPool.JOBS_SUBMIT):
        ctx.err = "job submission skipped (suite stopping)"
        ctx.ret_code = SuiteProcPool.JOB_SKIPPED_FLAG
        ctx.timestamp = get_current_time_string()
        return ctx

    try:
        stdin_file = None
        if ctx.cmd_kwargs.get('stdin_file_paths'):
            stdin_file = TemporaryFile()
            for file_path in ctx.cmd_kwargs['stdin_file_paths']:
                for line in open(file_path):
                    stdin_file.write(line)
            stdin_file.seek(0)
        elif ctx.cmd_kwargs.get('stdin_str'):
            stdin_file = PIPE
        proc = Popen(
            ctx.cmd, stdin=stdin_file, stdout=PIPE, stderr=PIPE,
            env=ctx.cmd_kwargs.get('env'), shell=ctx.cmd_kwargs.get('shell'))
    except IOError as exc:
        if cylc.flags.debug:
            traceback.print_exc()
        ctx.ret_code = 1
        ctx.err = str(exc)
    except OSError as exc:
        if exc.filename is None:
            exc.filename = ctx.cmd[0]
        if cylc.flags.debug:
            traceback.print_exc()
        ctx.ret_code = 1
        ctx.err = str(exc)
    else:
        ctx.out, ctx.err = proc.communicate(ctx.cmd_kwargs.get('stdin_str'))
        ctx.ret_code = proc.wait()

    ctx.timestamp = get_current_time_string()
    return ctx


class SuiteProcContext(object):
    """Represent the context of a command to run."""

    # Format string for single line output
    JOB_LOG_FMT_1 = "[%(cmd_key)s %(attr)s] %(mesg)s"
    # Format string for multi-line output
    JOB_LOG_FMT_M = "[%(cmd_key)s %(attr)s]\n%(mesg)s"

    def __init__(self, cmd_key, cmd, **cmd_kwargs):
        self.timestamp = get_current_time_string()
        self.cmd_key = cmd_key
        self.cmd = cmd
        self.cmd_kwargs = cmd_kwargs

        self.err = cmd_kwargs.get('err')
        self.ret_code = cmd_kwargs.get('ret_code')
        self.out = cmd_kwargs.get('out')

    def __str__(self):
        ret = ""
        for attr in "cmd", "ret_code", "out", "err":
            value = getattr(self, attr, None)
            if value is not None and str(value).strip():
                mesg = ""
                if attr == "cmd" and self.cmd_kwargs.get("stdin_file_paths"):
                    mesg += "cat"
                    for file_path in self.cmd_kwargs.get("stdin_file_paths"):
                        mesg += " " + quote(file_path)
                    mesg += " | "
                if attr == "cmd" and isinstance(value, list):
                    mesg += " ".join(quote(item) for item in value)
                else:
                    mesg = str(value).strip()
                if attr == "cmd" and self.cmd_kwargs.get("stdin_str"):
                    mesg += " <<<%s" % quote(self.cmd_kwargs.get("stdin_str"))
                if len(mesg.splitlines()) > 1:
                    fmt = self.JOB_LOG_FMT_M
                else:
                    fmt = self.JOB_LOG_FMT_1
                if not mesg.endswith("\n"):
                    mesg += "\n"
                ret += fmt % {
                    "cmd_key": self.cmd_key,
                    "attr": attr,
                    "mesg": mesg}
        return ret.rstrip()


class SuiteProcPool(object):
    """Use a process pool to execute shell commands."""

    JOBS_SUBMIT = "jobs-submit"
    JOB_SKIPPED_FLAG = 999
    # Shared memory flag.
    STOP_JOB_SUBMISSION = multiprocessing.Value('i', 0)

    _INSTANCE = None

    @classmethod
    def get_inst(cls, pool_size=None):
        """Return a singleton instance.

        On 1st call, instantiate the singleton. The argument "pool_size" is
        only relevant on 1st call.

        """
        if cls._INSTANCE is None:
            cls._INSTANCE = cls(pool_size)
        return cls._INSTANCE

    def __init__(self, pool_size=None):
        self.pool_size = (
            pool_size or
            GLOBAL_CFG.get(["process pool size"]) or
            multiprocessing.cpu_count())
        # (The Pool class defaults to cpu_count anyway, but does not
        # expose the result via its public interface).
        self.log = LOG
        self.log.debug(
            "Initializing process pool, size %d" % self.pool_size)
        self.pool = multiprocessing.Pool(processes=self.pool_size)
        self.results = {}

    def close(self):
        """Close the pool to new commands."""
        if not (self.is_dead() or self.is_closed()):
            self.log.debug("Closing process pool")
            self.pool.close()

    def handle_results_async(self):
        """Pass any available results to their associated callback."""
        for result_id, item in self.results.items():
            result, callback = item
            if result.ready():
                self.results.pop(result_id)
                value = result.get()
                if callable(callback):
                    callback(value)

    def is_closed(self):
        """Is the pool closed?"""
        # Warning: accesses multiprocessing.Pool internal state
        return self.pool._state == multiprocessing.pool.CLOSE

    def is_dead(self):
        """Have all my workers exited yet?"""
        # Warning: accesses multiprocessing.Pool internal state
        for pool in self.pool._pool:
            if pool.is_alive():
                return False
        return True

    def join(self):
        """Join after workers have exited. Close or terminate first."""
        self.log.debug("Joining process pool")
        self.pool.join()

    def put_command(self, ctx, callback):
        """Queue a new shell command to execute."""
        try:
            result = self.pool.apply_async(_run_command, [ctx])
        except AssertionError as exc:
            self.log.warning("%s\n  %s\n %s" % (
                str(exc),
                "Rejecting command (pool closed)",
                ctx.cmd))
        else:
            self.results[id(result)] = (result, callback)

    @staticmethod
    def run_command(ctx):
        """Execute a shell command and capture its output and exit status."""
        return _run_command(ctx)

    @classmethod
    def stop_job_submission(cls):
        """Set STOP_JOB_SUBMISSION flag."""
        cls.STOP_JOB_SUBMISSION.value = 1

    def terminate(self):
        """Kill all worker processes immediately."""
        if not self.is_dead():
            self.log.debug("Terminating process pool")
            self.pool.terminate()


def main():
    """Manual test playground."""

    LOG.setLevel(logging.INFO)  # or logging.DEBUG
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    LOG.addHandler(handler)

    def print_result(result):
        """Print result"""
        if result['OUT']:
            LOG.info('result> ' + result['OUT'].strip())
        if result['ERR']:
            LOG.info('FAILED> ' + result['CMD'])
            LOG.info(result['ERR'].strip())

    pool = mp_pool(3)

    for i in range(3):
        com = "sleep 5 && echo Hello from JOB " + str(i)
        pool.put_command(SuiteProcPool.JOBS_SUBMIT, com, print_result)
        com = "sleep 5 && echo Hello from POLL " + str(i)
        pool.put_command("poll", com, print_result)
        com = "sleep 5 && echo Hello from HANDLER " + str(i)
        pool.put_command("event-handler", com, print_result)
        com = "sleep 5 && echo Hello from HANDLER && badcommand"
        pool.put_command("event-handler", com, print_result)

    LOG.info('  sleeping')
    time.sleep(3)
    pool.handle_results_async()
    LOG.info('  sleeping')
    time.sleep(3)
    pool.close()
    # pool.terminate()
    pool.handle_results_async()
    LOG.info('  sleeping')
    time.sleep(3)
    pool.join()
    pool.handle_results_async()


if __name__ == '__main__':
    main()



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Task definition."""

from cylc.cycling.loader import get_point_relative, get_interval
from cylc.task_id import TaskID


class TaskDefError(Exception):
    """Exception raise for errors in TaskDef initialization."""

    def __init__(self, msg):
        Exception.__init__(self, msg)
        self.msg = msg

    def __str__(self):
        return "ERROR: %s" % self.msg


class TaskDef(object):
    """Task definition."""

    # Memory optimization - constrain possible attributes to this list.
    __slots__ = ["run_mode", "rtconfig", "start_point", "sequences",
                 "implicit_sequences", "used_in_offset_trigger",
                 "max_future_prereq_offset", "intercycle_offsets",
                 "sequential", "is_coldstart", "suite_polling_cfg",
                 "clocktrigger_offset", "expiration_offset",
                 "namespace_hierarchy", "triggers", "outputs",
                 "external_triggers", "name", "elapsed_times",
                 "mean_total_elapsed_time", "spawn_ahead"]

    def __init__(self, name, rtcfg, run_mode, start_point, spawn_ahead):
        if not TaskID.is_valid_name(name):
            raise TaskDefError("Illegal task name: %s" % name)

        self.run_mode = run_mode
        self.rtconfig = rtcfg
        self.start_point = start_point
        self.spawn_ahead = spawn_ahead

        self.sequences = []
        self.implicit_sequences = []  # Implicit sequences are deprecated.
        self.used_in_offset_trigger = False

        # some defaults
        self.max_future_prereq_offset = None
        self.intercycle_offsets = []
        self.sequential = False
        self.is_coldstart = False
        self.suite_polling_cfg = {}

        self.clocktrigger_offset = None
        self.expiration_offset = None
        self.namespace_hierarchy = []
        self.triggers = {}
        self.outputs = []

        self.external_triggers = []

        self.name = name
        self.elapsed_times = []
        self.mean_total_elapsed_time = None

    def add_trigger(self, triggers, expression, sequence):
        """Add triggers to a named sequence."""
        if sequence not in self.triggers:
            self.triggers[sequence] = []
        self.triggers[sequence].append([triggers, expression])

    def add_sequence(self, sequence, is_implicit=False):
        """Add a sequence."""
        if sequence not in self.sequences:
            self.sequences.append(sequence)
            if is_implicit:
                self.implicit_sequences.append(sequence)

    def describe(self):
        """Return title and description of the current task."""
        info = {}
        for item in 'title', 'description':
            info[item] = self.rtconfig[item]
        return info

    def check_for_explicit_cycling(self):
        """Check for explicitly somewhere.

        Must be called after all graph sequences added.
        """
        if len(self.sequences) == 0 and self.used_in_offset_trigger:
            raise TaskDefError(
                "No cycling sequences defined for %s" % self.name)

    @classmethod
    def get_cleanup_cutoff_point(cls, my_point, offset_sequence_tuples):
        """Extract the max dependent cycle point for this point."""
        if not offset_sequence_tuples:
            # This task does not have dependent tasks at other cycles.
            return my_point
        cutoff_points = []
        for offset_string, sequence in offset_sequence_tuples:
            if offset_string is None:
                # This indicates a dependency that lasts for the whole run.
                return None
            if sequence is None:
                # This indicates a simple offset interval such as [-PT6H].
                cutoff_points.append(
                    my_point - get_interval(offset_string))
                continue
            # This is a complicated offset like [02T00-P1W].
            dependent_point = sequence.get_start_point()

            matching_dependent_points = []
            while dependent_point is not None:
                # TODO: Is it realistically possible to hang in this loop?
                target_point = (
                    get_point_relative(offset_string, dependent_point))
                if target_point > my_point:
                    # Assume monotonic (target_point can never jump back).
                    break
                if target_point == my_point:
                    # We have found a dependent_point for my_point.
                    matching_dependent_points.append(dependent_point)
                dependent_point = sequence.get_next_point_on_sequence(
                    dependent_point)
            if matching_dependent_points:
                # Choose the largest of the dependent points.
                cutoff_points.append(matching_dependent_points[-1])
        if cutoff_points:
            max_cutoff_point = max(cutoff_points)
            if max_cutoff_point < my_point:
                # This is caused by future triggers - default to my_point.
                return my_point
            return max_cutoff_point
        # There aren't any dependent tasks in other cycles for my_point.
        return my_point

    def update_mean_total_elapsed_time(self, t_started, t_succeeded):
        """Update the mean total elapsed time (all instances of this task)."""
        if not t_started:
            # In case the started messaged did not come in.
            # (TODO - and we don't retain started time on restart?)
            return
        self.elapsed_times.append(t_succeeded - t_started)
        self.mean_total_elapsed_time = (
            sum(self.elapsed_times) / len(self.elapsed_times))



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

from cylc.cfgspec.globalcfg import GLOBAL_CFG


def prompt(question, force=False, gui=False, no_force=False, no_abort=False):
    """Interactive Yes/No prompt for cylc CLI scripts.

    For convenience, on No we just exit rather than return.
    If force is True don't prompt, just return immediately.

    """
    if (force or GLOBAL_CFG.get(['disable interactive command prompts'])) and (
            not no_force):
        return True
    if gui:
        import gtk
        dialog = gtk.MessageDialog(
            None, gtk.DIALOG_DESTROY_WITH_PARENT,
            gtk.MESSAGE_QUESTION, gtk.BUTTONS_YES_NO,
            question
        )
        gui_response = dialog.run()
        response_no = (gui_response != gtk.RESPONSE_YES)
    else:
        cli_response = raw_input('%s (y/n)? ' % question)
        response_no = (cli_response not in ['y', 'Y'])
    if response_no:
        if no_abort:
            return False
        else:
            sys.exit(0)
    else:
        return True



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Dump and load "cylc-suite-env" file."""

import os


class CylcSuiteEnvLoadError(Exception):
    """Error on loading cylc-suite-env file."""
    BASE_NAME = "cylc-suite-env"
    MISSING_KEY = (BASE_NAME + ": missing %(attr_key)s")
    UNEXPECTED_SUITE_NAME = (
        BASE_NAME + ": suite name expected [%(expected)s], got [%(actual)s]")


class CylcSuiteEnv(object):
    """Data structure, dumper and loader for suite contact file.

    An instance of this class has the following attributes:

    suite_name: the name of the suite
    suite_host: the name of the host on which the suite is running
    suite_port: the port number on host for connecting to the running suite
    suite_owner: the owner user ID of the suite
    suite_cylc_version: version of cylc for running the suite

    Constants:

    BASE_NAME: the default base name of the suite contact file
    ATTRS: map instance attributes to text keys in the suite contact file

    """

    BASE_NAME = CylcSuiteEnvLoadError.BASE_NAME
    ATTRS = {
        "CYLC_SUITE_NAME": "suite_name",
        "CYLC_SUITE_HOST": "suite_host",
        "CYLC_SUITE_PORT": "suite_port",
        "CYLC_SUITE_OWNER": "suite_owner",
        "CYLC_VERSION": "suite_cylc_version",
    }

    @classmethod
    def load(cls, suite_name, suite_run_dir):
        """Load suite contact file for suite_name from suite_run_dir.

        Return an instance of CylcSuiteEnv.
        Raise CylcSuiteEnvLoadError on failure.

        """
        ret = cls()
        try:
            for line in open(os.path.join(suite_run_dir, cls.BASE_NAME)):
                key, value = line.strip().split('=', 1)
                setattr(ret, cls.ATTRS[key], value)
        except (IOError, KeyError, ValueError):
            pass
        # Check that all expected attributes are defined
        for attr_key, attr in cls.ATTRS.items():
            if getattr(ret, attr) is None:
                raise CylcSuiteEnvLoadError(
                    CylcSuiteEnvLoadError.MISSING_KEY % {"attr_key": attr_key})
        # Check that we have loaded a file with the expected suite name
        if suite_name != ret.suite_name:
            raise CylcSuiteEnvLoadError(
                CylcSuiteEnvLoadError.UNEXPECTED_SUITE_NAME % {
                    "expected": suite_name, "actual": ret.suite_name})
        return ret

    def __init__(self, data=None):
        self.suite_name = None
        self.suite_host = None
        self.suite_port = None
        self.suite_owner = None
        self.suite_cylc_version = None
        if data:
            for attr_key, attr in self.ATTRS.items():
                setattr(self, attr, data.get(attr_key))

    def __str__(self):
        ret = ""
        for attr_key, attr in sorted(self.ATTRS.items()):
            ret += "%s=%s\n" % (attr_key, getattr(self, attr))
        return ret

    def dump(self, suite_run_dir):
        """Dump suite contact file under suite_run_dir.

        Don't bother staying alive on failure. Something must be very wrong.

        """
        with open(os.path.join(suite_run_dir, self.BASE_NAME), 'wb') as handle:
            handle.write(str(self))



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Task to cylc progress messaging."""

import os
import sys
from time import sleep
import traceback
from cylc.remote import remrun
from cylc.suite_env import CylcSuiteEnv, CylcSuiteEnvLoadError
from cylc.wallclock import get_current_time_string
import cylc.flags

from cylc.task_outputs import (
    TASK_OUTPUT_STARTED, TASK_OUTPUT_SUCCEEDED, TASK_OUTPUT_FAILED)


class TaskMessage(object):

    """Send task (job) output messages."""

    CYLC_JOB_PID = "CYLC_JOB_PID"
    CYLC_JOB_INIT_TIME = "CYLC_JOB_INIT_TIME"
    CYLC_JOB_EXIT = "CYLC_JOB_EXIT"
    CYLC_JOB_EXIT_TIME = "CYLC_JOB_EXIT_TIME"
    CYLC_MESSAGE = "CYLC_MESSAGE"

    FAIL_MESSAGE_PREFIX = "Task job script received signal "
    VACATION_MESSAGE_PREFIX = "Task job script vacated by signal "

    NORMAL = "NORMAL"
    WARNING = "WARNING"
    CRITICAL = "CRITICAL"
    PRIORITIES = (NORMAL, WARNING, CRITICAL)

    ATTRS = (
        ('suite', 'CYLC_SUITE_NAME', '(CYLC_SUITE_NAME)'),
        ('task_id', 'CYLC_TASK_ID', '(CYLC_TASK_ID)'),
        ('retry_seconds', 'CYLC_TASK_MSG_RETRY_INTVL',
         '(CYLC_TASK_MSG_RETRY_INTVL)'),
        ('max_tries', 'CYLC_TASK_MSG_MAX_TRIES', '(CYLC_TASK_MSG_MAX_TRIES)'),
        ('try_timeout', 'CYLC_TASK_MSG_TIMEOUT', '(CYLC_TASK_MSG_TIMEOUT)'),
        ('owner', 'CYLC_SUITE_OWNER', None),
        ('host', 'CYLC_SUITE_HOST', '(CYLC_SUITE_HOST)'),
        ('port', 'CYLC_SUITE_PORT', '(CYLC_SUITE_PORT)'),
    )

    def __init__(self, priority=NORMAL):
        if priority in self.PRIORITIES:
            self.priority = priority
        else:
            raise Exception('Illegal message priority ' + priority)

        # load the environment
        self.env_map = dict(os.environ)

        # set some instance variables
        self.suite = None
        self.task_id = None
        self.retry_seconds = None
        self.max_tries = None
        self.try_timeout = None
        self.owner = None
        self.host = None
        self.port = None
        for attr, key, default in self.ATTRS:
            value = self.env_map.get(key, default)
            setattr(self, attr, value)

        # conversions from string:
        if self.try_timeout == 'None':
            self.try_timeout = None
        try:
            self.retry_seconds = float(self.retry_seconds)
            self.max_tries = int(self.max_tries)
        except ValueError:
            pass

        # 'scheduler' or 'submit', (or 'raw' if job script run manually)
        self.mode = self.env_map.get('CYLC_MODE', 'raw')

        self.suite_run_dir = self.env_map.get('CYLC_SUITE_RUN_DIR', '.')

        self.utc = self.env_map.get('CYLC_UTC') == 'True'
        # Record the time the messaging system was called and append it
        # to the message, in case the message is delayed in some way.
        self.true_event_time = get_current_time_string(
            override_use_utc=self.utc)

        self.ssh_messaging = (
            self.env_map.get('CYLC_TASK_COMMS_METHOD') == 'ssh')

        self.polling = (
            self.env_map.get('CYLC_TASK_COMMS_METHOD') == 'poll')

        self.ssh_login_shell = (
            self.env_map.get('CYLC_TASK_SSH_LOGIN_SHELL') != 'False')

    def send(self, messages):
        """Send messages back to the suite."""
        self._update_job_status_file(messages)
        if self.mode != 'scheduler' or self.polling:
            # no suite to communicate with, just print to stdout.
            self._print_messages(messages)
            return

        if self.ssh_messaging and self._send_by_ssh():
            return

        self._send_by_pyro(messages)

    def _get_client(self):
        """Return the Pyro client."""
        from cylc.network.task_msgqueue import TaskMessageClient
        return TaskMessageClient(
            self.suite, self.task_id, self.owner, self.host,
            self.try_timeout, self.port)

    def _load_suite_contact_file(self):
        """Override CYLC_SUITE variables using the contact environment file.

        In case the suite was stopped and then restarted on another port.

        """
        try:
            suite_env = CylcSuiteEnv.load(self.suite, self.suite_run_dir)
        except CylcSuiteEnvLoadError:
            if cylc.flags.debug:
                traceback.print_exc()
        else:
            for key, attr_key in suite_env.ATTRS.items():
                self.env_map[key] = getattr(suite_env, attr_key)
        # set some instance variables
        for attr, key, default in (
                ('task_id', 'CYLC_TASK_ID', '(CYLC_TASK_ID)'),
                ('owner', 'CYLC_SUITE_OWNER', None),
                ('host', 'CYLC_SUITE_HOST', '(CYLC_SUITE_HOST)'),
                ('port', 'CYLC_SUITE_PORT', '(CYLC_SUITE_PORT)')):
            value = self.env_map.get(key, default)
            setattr(self, attr, value)

    def _print_messages(self, messages):
        """Print message to send."""
        prefix = 'cylc (%s - %s): ' % (self.mode, self.true_event_time)
        for message in messages:
            if self.priority == self.NORMAL:
                print prefix + message
            else:
                print >>sys.stderr, "%s%s %s" % (
                    prefix, self.priority, message)

    def _send_by_pyro(self, messages):
        """Send message by Pyro."""
        self._print_messages(messages)
        from Pyro.errors import NamingError
        sent = False
        i_try = 0
        while not sent and i_try < self.max_tries:
            i_try += 1
            try:
                # Get a proxy for the remote object and send the message.
                self._load_suite_contact_file()  # may change between tries
                client = self._get_client()
                for message in messages:
                    client.put(self.priority, message)
            except NamingError, exc:
                print >> sys.stderr, exc
                print "Send message: try %s of %s failed: %s" % (
                    i_try,
                    self.max_tries,
                    exc
                )
                print "Task proxy removed from suite daemon? Aborting."
                break
            except Exception, exc:
                print >> sys.stderr, exc
                print "Send message: try %s of %s failed: %s" % (
                    i_try,
                    self.max_tries,
                    exc
                )
                if i_try >= self.max_tries:
                    break
                print "   retry in %s seconds, timeout is %s" % (
                    self.retry_seconds,
                    self.try_timeout
                )
                sleep(self.retry_seconds)
            else:
                if i_try > 1:
                    print "Send message: try %s of %s succeeded" % (
                        i_try,
                        self.max_tries
                    )
                sent = True
        if not sent:
            # issue a warning and let the task carry on
            print >> sys.stderr, 'WARNING: MESSAGE SEND FAILED'

    def _send_by_ssh(self):
        """Send message via SSH."""
        self._load_suite_contact_file()

        # The suite definition specified that this task should
        # communicate back to the suite by means of using
        # non-interactive ssh to re-invoke the messaging command on the
        # suite host.

        # The remote_run() function expects command line options
        # to identify the target user and host names:
        sys.argv.append('--user=' + self.owner)
        sys.argv.append('--host=' + self.host)
        if cylc.flags.verbose:
            sys.argv.append('-v')

        if self.ssh_login_shell:
            sys.argv.append('--login')
        else:
            sys.argv.append('--no-login')

        # Some variables from the task execution environment are
        # also required by the re-invoked remote command: Note that
        # $CYLC_TASK_SSH_MESSAGING is not passed through so the
        # re-invoked command on the remote side will not end up in
        # this code block.
        env = {}
        for var in [
                'CYLC_MODE', 'CYLC_TASK_ID', 'CYLC_VERBOSE',
                'CYLC_SUITE_DEF_PATH_ON_SUITE_HOST',
                'CYLC_SUITE_RUN_DIR',
                'CYLC_SUITE_NAME', 'CYLC_SUITE_OWNER',
                'CYLC_SUITE_HOST', 'CYLC_SUITE_PORT', 'CYLC_UTC',
                'CYLC_TASK_MSG_MAX_TRIES', 'CYLC_TASK_MSG_TIMEOUT',
                'CYLC_TASK_MSG_RETRY_INTVL']:
            # (no exception handling here as these variables should
            # always be present in the task execution environment)
            env[var] = self.env_map.get(var, 'UNSET')

        # The path to cylc/bin on the remote end may be required:
        path = os.path.join(self.env_map['CYLC_DIR_ON_SUITE_HOST'], 'bin')

        # Return here if remote re-invocation occurred,
        # otherwise drop through to local Pyro messaging.
        # Note: do not sys.exit(0) here as the commands do, it
        # will cause messaging failures on the remote host.
        try:
            return remrun().execute(env=env, path=[path])
        except SystemExit:
            return

    def _update_job_status_file(self, messages):
        """Write messages to job status file."""
        job_log_name = os.getenv("CYLC_TASK_LOG_ROOT")
        job_status_file = None
        if job_log_name:
            try:
                job_status_file = open(job_log_name + ".status", "ab")
            except IOError as exc:
                if cylc.flags.debug:
                    print >>sys.stderr, exc
        for i, message in enumerate(messages):
            if job_status_file:
                if message == TASK_OUTPUT_STARTED:
                    job_status_file.write(
                        ("%s=%s\n" % (
                            self.CYLC_JOB_PID, os.getenv(self.CYLC_JOB_PID))) +
                        ("%s=%s\n" % (
                            self.CYLC_JOB_INIT_TIME, self.true_event_time)))
                elif message == TASK_OUTPUT_SUCCEEDED:
                    job_status_file.write(
                        ("%s=%s\n" % (self.CYLC_JOB_EXIT,
                                      TASK_OUTPUT_SUCCEEDED.upper())) +
                        ("%s=%s\n" % (
                            self.CYLC_JOB_EXIT_TIME, self.true_event_time)))
                elif message == TASK_OUTPUT_FAILED:
                    job_status_file.write("%s=%s\n" % (
                        self.CYLC_JOB_EXIT_TIME, self.true_event_time))
                elif message.startswith(self.FAIL_MESSAGE_PREFIX):
                    job_status_file.write("%s=%s\n" % (
                        self.CYLC_JOB_EXIT,
                        message.replace(self.FAIL_MESSAGE_PREFIX, "")))
                elif message.startswith(self.VACATION_MESSAGE_PREFIX):
                    # Job vacated, remove entries related to current job
                    job_status_file_name = job_status_file.name
                    job_status_file.close()
                    lines = []
                    for line in open(job_status_file_name):
                        if not line.startswith("CYLC_JOB_"):
                            lines.append(line)
                    job_status_file = open(job_status_file_name, "wb")
                    for line in lines:
                        job_status_file.write(line)
                    job_status_file.write("%s=%s|%s|%s\n" % (
                        self.CYLC_MESSAGE, self.true_event_time, self.priority,
                        message))
                else:
                    job_status_file.write("%s=%s|%s|%s\n" % (
                        self.CYLC_MESSAGE, self.true_event_time, self.priority,
                        message))
            messages[i] += ' at ' + self.true_event_time
        if job_status_file:
            try:
                job_status_file.close()
            except IOError as exc:
                if cylc.flags.debug:
                    print >>sys.stderr, exc



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Turn a cylc scheduler into a Unix daemon."""

import os
import sys
from cylc.suite_output import SuiteOutput


SUITE_SCAN_INFO_TMPL = r"""

To see if '%(suite)s' is running on '%(host)s:%(port)s':
 * cylc scan -n '\b%(suite)s\b' %(host)s
 * cylc ping -v --host=%(host)s %(suite)s
 * ssh %(host)s "pgrep -a -P 1 -fu $USER 'cylc-r.* \b%(suite)s\b'"

"""


_INFO_TMPL = r"""
Suite Info:
 + Name: %(suite)s
 + PID: %(pid)s
 + Host: %(host)s
 + Port: %(port)s
 + Logs: %(logd)s/{log,out,err}""" + SUITE_SCAN_INFO_TMPL


def daemonize(server):
    """Turn a cylc scheduler into a Unix daemon.

    Do the UNIX double-fork magic, see Stevens' "Advanced Programming in the
    UNIX Environment" for details (ISBN 0201563177)

    ATTRIBUTION: base on a public domain code recipe by Jurgen Hermann:
    http://code.activestate.com/recipes/66012-fork-a-daemon-process-on-unix/

    """

    sout = SuiteOutput(server.suite)

    try:
        pid = os.fork()
        if pid > 0:
            # exit first parent
            sys.exit(0)
    except OSError, exc:
        sys.stderr.write(
            "fork #1 failed: %d (%s)\n" % (exc.errno, exc.strerror))
        sys.exit(1)

    # decouple from parent environment
    os.chdir("/")
    os.setsid()
    os.umask(0)

    # do second fork
    try:
        pid = os.fork()
        if pid > 0:
            # exit from second parent, print eventual PID before
            sys.stdout.write(_INFO_TMPL % {
                "suite": server.suite,
                "host": server.host,
                "port": server.port,
                "pid": pid,
                "logd": os.path.dirname(sout.get_path())})
            sys.exit(0)
    except OSError, exc:
        sys.stderr.write(
            "fork #2 failed: %d (%s)\n" % (exc.errno, exc.strerror))
        sys.exit(1)

    # reset umask, octal
    os.umask(022)

    # redirect output to the suite log files
    sout.redirect()



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from cylc.batch_sys_handlers.background import BgCommandHandler


class MyBgCommandHandler(BgCommandHandler):

    """Job submission class for use by test battery.

    Allow a background submission to have a job vacation signal.

    """

    VACATION_SIGNAL = "USR1"

    def get_vacation_signal(self, _):
        return self.VACATION_SIGNAL


BATCH_SYS_HANDLER = MyBgCommandHandler()



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import unittest
from copy import copy
from task_id import TaskID
from parsec.OrderedDict import OrderedDictWithDefaults

"""Parameter expansion for runtime namespace names and graph strings.

Uses recursion to achieve nested looping over any number of parameters.  In its
simplest form (without allowing for parameter offsets and specific values, and
with input already expressed as a string template) the method looks like this:

#------------------------------------------------------------------------------
def expand(template, params, results, values=None):
    '''Recursive parameter expansion.

    template: e.g. "foo_m(m)s=>bar_m%(m)s_n%(n)s".
    results: output list of expanded strings.
    params: list of parameter (name, max-value) tuples.
    '''
    if values is None:
        values = {}
    if not params:
        results.add(template % values)
    else:
        param = params[0]
        for value in range(param[1]):
            values[param[0]] = value
            expand(template, params[1:], results, values)
#------------------------------------------------------------------------------
if __name__ == "__main__":
    results = []
    expand(
        "foo_m%(m)s=>bar_m%(m)s_n%(n)s",
        results,
        [('m', 2), ('n', 3)]
    )
    for result in results:
        print result

foo_m0=>bar_m0_n0
foo_m0=>bar_m0_n1
foo_m0=>bar_m0_n2
foo_m1=>bar_m1_n0
foo_m1=>bar_m1_n1
foo_m1=>bar_m1_n2
#------------------------------------------------------------------------------
"""

# To split runtime heading name lists.
REC_NAMES = re.compile(r'(?:[^,<]|\<[^>]*\>)+')
# To extract 'name', '<parameters>', and 'other' from
#   'name<parameters>other' (other is used for clock-offsets).
REC_P_NAME = re.compile(r"(%s)(<.*?>)?(.+)?" % TaskID.NAME_RE)
# To extract all parameter lists e.g. 'm,n,o' (from '<m,n,o>').
REC_P_GROUP = re.compile(r"<(.*?)>")
# To extract parameter name and optional offset or value e.g. 'm-1'.
REC_P_OFFS = re.compile(r'(\w+)(?:\s*([-+=]\s*[\w]+))?')


def item_in_list(item, lst):
    """Return True if item is in lst, by string or int comparison.

    Items may be general strings, or strings of zero-padded integers.
    """
    if item in lst:
        return True
    try:
        int(item)
    except ValueError:
        return False
    return int(item) in [int(i) for i in lst]


class ParamExpandError(Exception):
    """For parameter expansion errors."""
    pass


class NameExpander(object):
    """Handle parameter expansion in runtime namespace headings."""

    def __init__(self, parameters=None):
        """Initialize the parameterized task name expander.

        parameters (None if no parameters are defined) is:
            ({param_name: [param_values],  # list of strings
             {param_name: param_template}) # e.g. "_m%(m)s"
        """
        self.parameters = parameters or ({}, {})
        self.param_cfg = self.parameters[0]
        self.param_tmpl_cfg = self.parameters[1]

    def expand(self, runtime_heading):
        """Expand runtime namespace names for a subset of suite parameters.

        Input runtime_heading is a string that may contain comma-separated
        parameterized namespace names, e.g. for "foo<m,n>, bar<m,n>".

        Unlike GraphExpander this does not support offsets like "foo<m-1,n>",
        but it does support specific parameter values like "foo<m=0,n>".

        Returns a list of tuples, each with an expanded name and its parameter
        values (to be passed to the corresponding tasks), e.g.:
            [('foo_i0_j0', {i:'0', j:'0'}),
             ('foo_i0_j1', {i:'0', j:'1'}),
             ('foo_i1_j0', {i:'1', j:'0'}),
             ('foo_i1_j1', {i:'1', j:'1'})]
        """
        # Create a string template and values to pass to the expansion method.
        expanded = []
        for namespace in REC_NAMES.findall(runtime_heading):
            template = namespace.strip()
            name, p_tmpl, other = REC_P_NAME.match(template).groups()
            if not p_tmpl:
                # Not parameterized.
                if other:
                    expanded.append((name + other, {}))
                else:
                    expanded.append((name, {}))
                continue
            tmpl = name
            # Get the subset of parameters used in this case.
            used_param_names = []
            spec_vals = {}
            for item in p_tmpl[1:-1].split(','):
                pname, sval = REC_P_OFFS.match(item.strip()).groups()
                if pname not in self.param_cfg:
                    raise ParamExpandError(
                        "ERROR, parameter %s is not defined in %s" % (
                            pname, runtime_heading))
                if sval:
                    if sval.startswith('+') or sval.startswith('-'):
                        raise ParamExpandError(
                            "ERROR, parameter index offsets are not"
                            " supported in name expansion: %s%s" % (
                                pname, sval))
                    elif sval.startswith('='):
                        # Check that specific parameter values exist.
                        val = sval[1:].strip()
                        # Pad integer values here.
                        try:
                            int(val)
                        except ValueError:
                            nval = val
                        else:
                            nval = val.zfill(len(self.param_cfg[pname][0]))
                        if not item_in_list(nval, self.param_cfg[pname]):
                            raise ParamExpandError(
                                "ERROR, parameter %s out of range: %s" % (
                                    pname, p_tmpl))
                        spec_vals[pname] = nval
                else:
                    used_param_names.append(pname)
                tmpl += self.param_tmpl_cfg[pname]
            if other:
                tmpl += other
            used_params = [
                (p, self.param_cfg[p]) for p in used_param_names]
            self._expand_name(tmpl, used_params, expanded, spec_vals)
        return expanded

    def _expand_name(self, str_tmpl, param_list, results, spec_vals=None):
        """Expand str_tmpl for any number of parameters.

        str_tmpl is a string template, e.g. 'foo_m%(m)s_n%(n)s' for two
            parameters m and n.
        param_list is a list of tuples (name, max-val) for each parameter
            to be looped over.
        spec_vals is a map of values for parameters that are not to be looped
            over because they've been assigned a specific value.

        E.g. for "foo<m=0,n>" str_tmpl is "foo_m%(m)s_n%(n)s", param_list is
        [('n', 2)], and spec_values {'m': 0}.

        results contains the expanded names and corresponding parameter values,
        as described above in the calling method.
        """
        if spec_vals is None:
            spec_vals = {}
        if not param_list:
            # Inner loop.
            current_values = copy(spec_vals)
            results.append((str_tmpl % current_values, current_values))
        else:
            for param_val in param_list[0][1]:
                spec_vals[param_list[0][0]] = param_val
                self._expand_name(str_tmpl, param_list[1:], results, spec_vals)

    def replace_params(self, name_in, param_values, origin):
        """Replace parameters in name_in with values in param_values.

        Note this is "expansion" for specific values, not all values.
        """
        name, p_tmpl, _ = REC_P_NAME.match(name_in).groups()
        if not p_tmpl:
            # name_in is not parameterized.
            return name_in
        # List of parameter names used in this name: ['m', 'n']
        used_param_names = [i.strip() for i in p_tmpl[1:-1].split(',')]
        for p_name in used_param_names:
            msg = None
            if '=' in p_name:
                msg = 'values'
            elif '-' in p_name or '+' in p_name:
                msg = 'offsets'
            if msg is not None:
                raise ParamExpandError("ERROR, parameter %s not supported"
                                       " here: %s" % (msg, origin))
        str_template = name
        for pname in used_param_names:
            str_template += self.param_tmpl_cfg[pname]
        return str_template % param_values


class GraphExpander(object):
    """Handle parameter expansion of graph string lines."""

    def __init__(self, parameters=None):
        """Initialize the parameterized task name expander.

        parameters (None if no parameters are defined) is:
            ({param_name: [param_values],  # list of strings
             {param_name: param_template}) # e.g. "_m%(m)s"
        """
        self.parameters = parameters or ({}, {})
        self.param_cfg = self.parameters[0]
        self.param_tmpl_cfg = self.parameters[1]

    def expand(self, line):
        """Expand a graph line for subset of suite parameters.

        Input line is a string that may contain multiple parameterized node
        names, e.g. "pre=>init<m>=>sim<m,n>=>post<m,n>=>done".

        Unlike NameExpander this supports offsets like "foo<m-1,n>", which
        means (because the parameter substitutions have to be computed on the
        fly) we have shift creation of the expansion string template into the
        inner loop of the recursive expansion function.

        Returns a set containing lines expanded for all used parameters, e.g.
        for "foo=>bar<m,n>" with m=2 and n=2 the result would be:
            set([foo=>bar_m0_n0,
                 foo=>bar_m0_n1,
                 foo=>bar_m1_n0,
                 foo=>bar_m1_n1])

        Specific parameter values can be singled out like this:
            "sim<m=0,n>=>sim<m,n>"
        Offset (negative only) values can be specified like this:
            "sim<m-1,n>=>sim<m,n>"
        (Here the offset node must be the first in a line, and if m-1 evaluates
        to less than 0 the node will be removed to leave just "sim<m,n>").
        """
        line_set = set()
        used_pnames = []
        for p_group in set(REC_P_GROUP.findall(line)):
            for item in p_group.split(','):
                pname, offs = REC_P_OFFS.match(item).groups()
                if pname not in self.param_cfg:
                    raise ParamExpandError(
                        "ERROR, parameter %s is not defined in <%s>: %s" % (
                            pname, p_group, line))
                if offs:
                    if offs.startswith('+'):
                        raise ParamExpandError(
                            "ERROR, +ve parameter offsets are not"
                            " supported: %s%s" % (pname, offs))
                    elif offs.startswith('='):
                        # Check that specific parameter values exist.
                        val = offs[1:]
                        # Pad integer values here.
                        try:
                            int(val)
                        except ValueError:
                            nval = val
                        else:
                            nval = val.zfill(len(self.param_cfg[pname][0]))
                            if nval != val:
                                line = re.sub(item,
                                              '%s=%s' % (pname, nval), line)
                        if not item_in_list(nval, self.param_cfg[pname]):
                            raise ParamExpandError(
                                "ERROR, parameter %s out of range: %s" % (
                                    pname, p_group))
                if pname not in used_pnames:
                    used_pnames.append(pname)
        used_params = [(p, self.param_cfg[p]) for p in used_pnames]
        self._expand_graph(line, dict(used_params), used_params, line_set)
        return line_set

    def _expand_graph(self, line, all_params,
                      param_list, line_set, values=None):
        """Expand line into line_set for any number of parameters.

        line is a graph string line as described above in the calling method.
        param_list is a list of tuples (name, max-val) for each parameter.
        results is a set to hold each expanded line.
        """
        if values is None:
            values = {}
        if not param_list:
            # Inner loop.
            for p_group in set(REC_P_GROUP.findall(line)):
                # Parameters must be expanded in the order found.
                param_values = OrderedDictWithDefaults()
                tmpl = ""
                for item in p_group.split(','):
                    pname, offs = REC_P_OFFS.match(item).groups()
                    if offs is None:
                        param_values[pname] = values[pname]
                    elif offs.startswith('='):
                        # Specific value.
                        param_values[pname] = offs[1:]
                    else:
                        # Index offset.
                        plist = all_params[pname]
                        cur_idx = plist.index(values[pname])
                        off_idx = cur_idx + int(offs)
                        if off_idx < 0:
                            offval = "--<REMOVE>--"
                        else:
                            offval = plist[off_idx]
                        param_values[pname] = offval
                for pname in param_values:
                    tmpl += self.param_tmpl_cfg[pname]
                repl = tmpl % param_values
                line = re.sub('<' + p_group + '>', repl, line)
                # Remove out-of-range nodes to first arrow.
                line = re.sub('^.*--<REMOVE>--.*?=>\s*?', '', line)
            line_set.add(line)
        else:
            # Recurse through index ranges.
            for param_val in param_list[0][1]:
                values[param_list[0][0]] = param_val
                self._expand_graph(line, all_params,
                                   param_list[1:], line_set, values)


class TestParamExpand(unittest.TestCase):
    """Unit tests for the parameter expansion module."""

    def setUp(self):
        """Create some parameters and templates for use in tests."""
        ivals = [str(i) for i in range(2)]
        jvals = [str(j) for j in range(3)]
        kvals = [str(k) for k in range(2)]
        params_map = {'i': ivals, 'j': jvals, 'k': kvals}
        templates = {'i': '_i%(i)s',
                     'j': '_j%(j)s',
                     'k': '_k%(k)s'}
        self.name_expander = NameExpander((params_map, templates))
        self.graph_expander = GraphExpander((params_map, templates))

    def test_name_one_param(self):
        """Test name expansion and returned value for a single parameter."""
        self.assertEqual(
            self.name_expander.expand('foo<j>'),
            [('foo_j0', {'j': '0'}),
             ('foo_j1', {'j': '1'}),
             ('foo_j2', {'j': '2'})]
        )

    def test_name_two_params(self):
        """Test name expansion and returned values for two parameters."""
        self.assertEqual(
            self.name_expander.expand('foo<i,j>'),
            [('foo_i0_j0', {'i': '0', 'j': '0'}),
             ('foo_i0_j1', {'i': '0', 'j': '1'}),
             ('foo_i0_j2', {'i': '0', 'j': '2'}),
             ('foo_i1_j0', {'i': '1', 'j': '0'}),
             ('foo_i1_j1', {'i': '1', 'j': '1'}),
             ('foo_i1_j2', {'i': '1', 'j': '2'})]
        )

    def test_name_two_names(self):
        """Test name expansion for two names."""
        self.assertEqual(
            self.name_expander.expand('foo<i>, bar<j>'),
            [('foo_i0', {'i': '0'}),
             ('foo_i1', {'i': '1'}),
             ('bar_j0', {'j': '0'}),
             ('bar_j1', {'j': '1'}),
             ('bar_j2', {'j': '2'})]
        )

    def test_name_specific_val_1(self):
        """Test singling out a specific value, in name expansion."""
        self.assertEqual(
            self.name_expander.expand('foo<i=0>'),
            [('foo_i0', {'i': '0'})]
        )

    def test_name_specific_val_2(self):
        """Test specific value in the first parameter of a pair."""
        self.assertEqual(
            self.name_expander.expand('foo<i=0,j>'),
            [('foo_i0_j0', {'i': '0', 'j': '0'}),
             ('foo_i0_j1', {'i': '0', 'j': '1'}),
             ('foo_i0_j2', {'i': '0', 'j': '2'})]
        )

    def test_name_specific_val_3(self):
        """Test specific value in the second parameter of a pair."""
        self.assertEqual(
            self.name_expander.expand('foo<i,j=1>'),
            [('foo_i0_j1', {'i': '0', 'j': '1'}),
             ('foo_i1_j1', {'i': '1', 'j': '1'})]
        )

    def test_name_fail_bare_value(self):
        """Test foo<0,j> fails."""
        # It should be foo<i=0,j>.
        self.assertRaises(ParamExpandError,
                          self.name_expander.expand,
                          'foo<0,j>')

    def test_name_fail_undefined_param(self):
        """Test that an undefined parameter gets failed."""
        # m is not defined.
        self.assertRaises(ParamExpandError,
                          self.name_expander.expand,
                          'foo<m,j>')

    def test_name_fail_param_value_too_high(self):
        """Test that an out-of-range parameter gets failed."""
        # i stops at 3.
        self.assertRaises(ParamExpandError,
                          self.name_expander.expand,
                          'foo<i=4,j>')

    def test_name_multiple(self):
        """Test expansion of two names, with one and two parameters."""
        self.assertEqual(
            self.name_expander.expand('foo<i>, bar<i,j>'),
            [('foo_i0', {'i': '0'}),
             ('foo_i1', {'i': '1'}),
             ('bar_i0_j0', {'i': '0', 'j': '0'}),
             ('bar_i0_j1', {'i': '0', 'j': '1'}),
             ('bar_i0_j2', {'i': '0', 'j': '2'}),
             ('bar_i1_j0', {'i': '1', 'j': '0'}),
             ('bar_i1_j1', {'i': '1', 'j': '1'}),
             ('bar_i1_j2', {'i': '1', 'j': '2'})]
        )

    def test_graph_expand_1(self):
        """Test graph expansion with two parameters each side of an arrow."""
        self.assertEqual(
            self.graph_expander.expand("bar<i,j>=>baz<i,j>"),
            set(["bar_i0_j1=>baz_i0_j1",
                 "bar_i1_j2=>baz_i1_j2",
                 "bar_i0_j2=>baz_i0_j2",
                 "bar_i1_j1=>baz_i1_j1",
                 "bar_i1_j0=>baz_i1_j0",
                 "bar_i0_j0=>baz_i0_j0"])
        )

    def test_graph_expand_2(self):
        """Test graph expansion to 'branch and merge' a workflow."""
        self.assertEqual(
            self.graph_expander.expand(
                "pre=>bar<i>=>baz<i,j>=>post"),
            set(["pre=>bar_i0=>baz_i0_j1=>post",
                 "pre=>bar_i1=>baz_i1_j2=>post",
                 "pre=>bar_i0=>baz_i0_j2=>post",
                 "pre=>bar_i1=>baz_i1_j1=>post",
                 "pre=>bar_i1=>baz_i1_j0=>post",
                 "pre=>bar_i0=>baz_i0_j0=>post"])
        )

    def test_graph_expand_offset(self):
        """Test graph expansion with an offset."""
        self.assertEqual(
            self.graph_expander.expand(
                "bar<i-1,j>=>baz<i,j>"),
            set(["baz_i0_j0",
                 "baz_i0_j1",
                 "baz_i0_j2",
                 "bar_i0_j0=>baz_i1_j0",
                 "bar_i0_j1=>baz_i1_j1",
                 "bar_i0_j2=>baz_i1_j2"])
        )

    def test_graph_expand_specific(self):
        """Test graph expansion with a specific value."""
        self.assertEqual(
            self.graph_expander.expand("bar<i=1,j>=>baz<i,j>"),
            set(["bar_i1_j0=>baz_i0_j0",
                 "bar_i1_j1=>baz_i0_j1",
                 "bar_i1_j2=>baz_i0_j2",
                 "bar_i1_j0=>baz_i1_j0",
                 "bar_i1_j1=>baz_i1_j1",
                 "bar_i1_j2=>baz_i1_j2"])
        )

    def test_graph_fail_bare_value(self):
        """Test that a bare parameter value fails in the graph."""
        self.assertRaises(ParamExpandError,
                          self.graph_expander.expand,
                          'foo<0,j>=>bar<i,j>')

    def test_graph_fail_undefined_param(self):
        """Test that an undefined parameter value fails in the graph."""
        self.assertRaises(ParamExpandError,
                          self.graph_expander.expand,
                          'foo<m,j>=>bar<i,j>')

    def test_graph_fail_param_value_too_high(self):
        """Test that an out-of-range parameter value fails in the graph."""
        self.assertRaises(ParamExpandError,
                          self.graph_expander.expand,
                          'foo<i=4,j><i,j>')


if __name__ == "__main__":
    unittest.main()



#############################################################################
#  
#	Event Service client base classes
#
#	This is part of "Pyro" - Python Remote Objects
#	which is (c) Irmen de Jong - irmen@razorvine.net
#
#############################################################################

import Pyro.core, Pyro.naming, Pyro.constants
import Pyro.EventService.Server
from Pyro.EventService.Event import Event
from Pyro.errors import *

# SUBSCRIBER: subscribes to certain events.
class Subscriber(Pyro.core.CallbackObjBase):
	def __init__(self, ident=None, esURI=None):
		Pyro.core.CallbackObjBase.__init__(self)
		Pyro.core.initServer()
		Pyro.core.initClient()
		daemon = Pyro.core.Daemon()
		if esURI:
			check=Pyro.core.PyroURI(esURI)
			self.ES_uri=esURI
		else:		
			locator = Pyro.naming.NameServerLocator(identification=ident)
			self.NS = locator.getNS(host=Pyro.config.PYRO_NS_HOSTNAME)
			daemon.useNameServer(self.NS)
			self.ES_uri = self.NS.resolve(Pyro.constants.EVENTSERVER_NAME)
		daemon.connect(self)  #  will also set self.daemon...
		self.ES_ident=ident
		self.abortListen=0
		self.daemon=daemon	# make sure daemon doesn't get garbage collected now

	def getES(self):
		# we get a fresh proxy to the ES because of threading issues.
		# (proxies can not be reused across multiple threads)
		eventservice=Pyro.core.getProxyForURI(self.ES_uri)
		eventservice._setIdentification(self.ES_ident)
		return eventservice

	def subscribe(self,subjects):
		# Subscribe to one or more subjects.
		# It is safe to call this multiple times.
		self.getES().subscribe(subjects, self.getProxy())
	def subscribeMatch(self,subjectPatterns):
		# Subscribe to one or more subjects (by pattern)
		# It is safe to call this multiple times.
		self.getES().subscribeMatch(subjectPatterns, self.getProxy())
	def unsubscribe(self, subjects):
		# Unsubscribe the subscriber for the given subject(s).
		self.getES().unsubscribe(subjects, self.getProxy())

	def abort(self):
		self.abortListen=1

	def setThreading(self, threaded):
		self.getDaemon().threaded=threaded

	def listen(self):
		self.getDaemon().requestLoop(lambda s=self: not s.abortListen)

	def event(self, event):					# callback, override this!
		print event

# PUBLISHER: publishes events.
class Publisher(object):
	def __init__(self, ident=None, esURI=None):
		Pyro.core.initClient()
		if esURI:
			check=Pyro.core.PyroURI(esURI)
			self.ES_uri=esURI
		else:
			locator = Pyro.naming.NameServerLocator(identification=ident)
			ns = locator.getNS(host=Pyro.config.PYRO_NS_HOSTNAME)
			self.ES_uri = ns.resolve(Pyro.constants.EVENTSERVER_NAME)
			ns._release()   # be very sure to release the socket
		self.ES_ident = ident

	def getES(self):
		# we get a fresh proxy to the ES because of threading issues.
		# (proxies can not be reused across multiple threads)
		eventservice=Pyro.core.getProxyForURI(self.ES_uri)
		eventservice._setIdentification(self.ES_ident)
		return eventservice

	def publish(self, subjects, msg):
		es=self.getES()
		es.publish(subjects,msg)
		es._release()   # be very sure to release the socket




#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Host name utilities

ATTRIBUTION:
http://www.linux-support.com/cms/get-local-ip-address-with-python/

Fetching the outgoing IP address of a computer might be a difficult
task. Computers can contain a large set of network devices, each
connected to different and independent sub-networks. Additionally there
might be available a number of devices, to be utilized in the manner of
network devices to exchange data with external systems.

However, if properly configured, your operating system knows what device
has to be utilized. Querying results depend on target addresses and
routing information. In our solution we are utilizing the features of
the local operating system to determine the correct network device. It is
the same step we will get the associated network address.

To reach this goal we will utilize the UDP protocol. Unlike TCP/IP, UDP
is a stateless networking protocol to transfer single data packages. You
do not have to open a point-to-point connection to a service running at
the target host. We have to provide the target address to enable the
operating system to find the correct device. Due to the nature of UDP
you are not required to choose a valid target address. You just have to
make sure your are choosing an arbitrary address from the correct
subnet.

The following function is temporarily opening a UDP server socket. It is
returning the IP address associated with this socket.

"""

import sys
import socket
from time import time


class SuiteHostUtil(object):
    """(Suite) host utility."""

    EXPIRE = 3600.0  # singleton expires in 1 hour by default
    _instance = None

    @classmethod
    def get_inst(cls, new=False, expire=None):
        """Return the singleton instance of this class.

        "new": if True, create a new singleton instance.
        "expire":
            the expire duration in seconds. If None or not specified, the
            singleton expires after 3600.0 seconds (1 hour). Once expired, the
            next call to this method will create a new singleton.

        """
        if expire is None:
            expire = cls.EXPIRE
        if cls._instance is None or new or time() > cls._instance.expire_time:
            cls._instance = cls(expire)
        return cls._instance

    def __init__(self, expire):
        self.expire_time = time() + expire
        self._host_name = None
        self._host_ip_address = None
        self._host_name_pref = None
        self._remote_hosts = {}  # host: is_remote, ...

    @staticmethod
    def get_local_ip_address(target):
        """Return IP address of target.

        This finds the external address of the particular network adapter
        responsible for connecting to the target?

        Note that although no connection is made to the target, the target
        must be reachable on the network (or just recorded in the DNS?) or
        the function will hang and time out after a few seconds.

        """

        ipaddr = ""
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.connect((target, 8000))
            ipaddr = sock.getsockname()[0]
            sock.close()
        except IOError:
            pass
        return ipaddr

    def get_hostname(self):
        """Return the fully qualified domain name for current host."""
        if self._host_name is None:
            self._host_name = socket.getfqdn()
        return self._host_name

    def _get_host_ip_address(self):
        """Return the external IP address for the current host."""
        if self._host_ip_address is None:
            self._host_ip_address = self.get_local_ip_address(
                self._get_identification_cfg('target'))
        return self._host_ip_address

    @staticmethod
    def _get_identification_cfg(key):
        """Return the [suite host self-identification]key global conf."""
        from cylc.cfgspec.globalcfg import GLOBAL_CFG
        return GLOBAL_CFG.get(['suite host self-identification', key])

    def get_suite_host(self):
        """Return the preferred identifier for the suite (or current) host.

        As specified by the "suite host self-identification" settings in the
        site/user global.rc files. This is mainly used for suite host
        identification by task jobs.

        """
        if self._host_name_pref is None:
            hardwired = self._get_identification_cfg('host')
            method = self._get_identification_cfg('method')
            if method == 'address':
                self._host_name_pref = self._get_host_ip_address()
            elif method == 'hardwired' and hardwired:
                self._host_name_pref = hardwired
            else:  # if method == 'name':
                self._host_name_pref = self.get_hostname()
        return self._host_name_pref

    def is_remote_host(self, name):
        """Return True if name has different IP address than the current host.

        Return False if name is None.
        Raise IOError if host is unknown.

        """
        if name not in self._remote_hosts:
            if not name or name.split(".")[0].startswith("localhost"):
                # e.g. localhost.localdomain
                self._remote_hosts[name] = False
            else:
                try:
                    ipa = socket.gethostbyname(name)
                except IOError as exc:
                    if exc.filename is None:
                        exc.filename = str(name)
                    raise
                host_ip_address = self._get_host_ip_address()
                # local IP address of the suite host (e.g. 127.0.0.1)
                local_ip_address = socket.gethostbyname(self.get_hostname())
                self._remote_hosts[name] = (
                    ipa not in [host_ip_address, local_ip_address])
        return self._remote_hosts[name]


def get_hostname():
    """Shorthand for SuiteHostUtil.get_inst().get_hostname()."""
    return SuiteHostUtil.get_inst().get_hostname()


def get_local_ip_address(target):
    """Shorthand for SuiteHostUtil.get_inst().get_local_ip_address(target)."""
    return SuiteHostUtil.get_inst().get_local_ip_address(target)


def get_suite_host():
    """Shorthand for SuiteHostUtil.get_inst().get_suite_host()."""
    return SuiteHostUtil.get_inst().get_suite_host()


def is_remote_host(name):
    """Shorthand for SuiteHostUtil.get_inst().is_remote_host(name)."""
    return SuiteHostUtil.get_inst().is_remote_host(name)


if __name__ == "__main__":
    sys.stdout.write("%s\n" % get_local_ip_address(sys.argv[1]))



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# A minimal Pyro-connected object to allow client programs to identify
# what suite is running at a given cylc port - by suite name and owner.

# All *other* suite objects should be connected to Pyro via qualified
# names: owner.suite.object, to prevent accidental access to the wrong
# suite. This object, however, should be connected unqualified so that
# that same ID method can be called on any active cylc port.

import cylc.flags
from cylc.network.pyro_base import PyroServer
from cylc.network.suite_state import StateSummaryServer
from cylc.network import access_priv_ok
from cylc.config import SuiteConfig


class SuiteIdServer(PyroServer):
    """Server-side external trigger interface."""

    _INSTANCE = None

    @classmethod
    def get_inst(cls, name=None, owner=None):
        """Return a singleton instance."""
        if cls._INSTANCE is None:
            cls._INSTANCE = cls(name, owner)
        return cls._INSTANCE

    def __init__(self, name, owner):
        self.owner = owner
        self.name = name
        super(SuiteIdServer, self).__init__()

    def identify(self):
        self.report("identify")
        result = {}
        if access_priv_ok(self, "identity"):
            result['name'] = self.name
            result['owner'] = self.owner
        if access_priv_ok(self, "description"):
            config = SuiteConfig.get_inst()
            result['title'] = config.cfg['title']
            result['description'] = config.cfg['description']
        if access_priv_ok(self, "state-totals"):
            result['states'] = StateSummaryServer.get_inst().get_state_totals()
            result['update-time'] = (
                StateSummaryServer.get_inst().get_summary_update_time())
            result['tasks-by-state'] = (
                StateSummaryServer.get_inst().get_tasks_by_state())
        return result

    def id(self):
        # Back-compat for older clients <=6.4.1.
        # (Allows old scan to see new suites.)
        return (self.name, self.owner)



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gobject
import gtk
import threading
from time import sleep

from cylc.task_id import TaskID
from cylc.gui.dot_maker import DotMaker
from cylc.network.suite_state import get_id_summary
from copy import deepcopy

import warnings
warnings.filterwarnings('ignore', '^.*was not found when attempting to ' +
                        'remove it', Warning)


class DotUpdater(threading.Thread):

    RIGHT_ARROW = u'\u25b7'  # Unicode enpty triangle facing right.
    DOWN_ARROW = u'\u25bd'  # Unicode enpty triangle facing down.
    INDENT = 5  # Number of spaces to indent nested items in transpose view.

    def __init__(self, cfg, updater, treeview, info_bar, theme, dot_size):

        super(DotUpdater, self).__init__()

        self.quit = False
        self.cleared = True
        self.action_required = False
        self.autoexpand = True
        self.should_hide_headings = False
        self.should_group_families = ("dot" not in cfg.ungrouped_views)
        self.should_transpose_view = False
        self.is_transposed = False
        self.defn_order_on = True

        self.cfg = cfg
        self.updater = updater
        self.theme = theme
        self.info_bar = info_bar
        self.last_update_time = None
        self.state_summary = {}
        self.fam_state_summary = {}
        self.ancestors_pruned = {}
        self.descendants = []
        self.point_strings = []

        self.led_headings = []
        self.led_treeview = treeview
        self.led_treestore = treeview.get_model()
        self._prev_tooltip_task_id = None
        if hasattr(self.led_treeview, "set_has_tooltip"):
            self.led_treeview.set_has_tooltip(True)
            try:
                self.led_treeview.connect('query-tooltip',
                                          self.on_query_tooltip)
            except TypeError:
                # Lower PyGTK version.
                pass

        self.task_list = []
        self.family_tree = {}
        self.expanded_rows = []

        # generate task state icons
        dotm = DotMaker(theme, size=dot_size)
        self.dots = dotm.get_dots()

    def _set_tooltip(self, widget, tip_text):
        tip = gtk.Tooltips()
        tip.enable()
        tip.set_tip(widget, tip_text)

    def clear_list(self):
        self.led_treestore.clear()
        # gtk idle functions must return false or will be called multiple times
        return False

    def update(self):
        if not self.updater.connected:
            if not self.cleared:
                gobject.idle_add(self.clear_list)
                self.cleared = True
            return False
        self.cleared = False

        if not self.action_required and (
                self.last_update_time is not None and
                self.last_update_time >= self.updater.last_update_time):
            return False

        self.last_update_time = self.updater.last_update_time
        self.updater.set_update(False)

        self.state_summary = deepcopy(self.updater.state_summary)
        self.fam_state_summary = deepcopy(self.updater.fam_state_summary)
        self.ancestors_pruned = deepcopy(self.updater.ancestors_pruned)
        self.descendants = deepcopy(self.updater.descendants)

        self.updater.set_update(True)

        self.point_strings = []
        for id_ in self.state_summary:
            name, point_string = TaskID.split(id_)
            if point_string not in self.point_strings:
                self.point_strings.append(point_string)
        try:
            self.point_strings.sort(key=int)
        except (TypeError, ValueError):
            # iso cycle points
            self.point_strings.sort()

        use_def_order = (
            self.cfg.use_defn_order and self.updater.ns_defn_order and
            self.defn_order_on)

        if not self.should_group_families:
            # Display the full task list.
            self.task_list = deepcopy(self.updater.task_list)

            if use_def_order:
                self.task_list = [task for task in self.updater.ns_defn_order
                                  if task in self.task_list]
            else:
                self.task_list.sort()
        else:
            self.family_tree = {}
            self.task_list = deepcopy(self.updater.task_list)

            if use_def_order:
                self.task_list = [task for task in self.updater.ns_defn_order
                                  if task in self.task_list]
            else:
                self.task_list.sort()

            for task in self.task_list:
                # Generate dict of families and their associated tasks.
                item = self.ancestors_pruned[task][-2]
                if item not in self.task_list:
                    if item not in self.family_tree:
                        self.family_tree[item] = []
                    self.family_tree[item].append(task)

            for heading, tasks in self.family_tree.iteritems():
                # Place associated tasks after headers.
                ind = min([self.task_list.index(task) for task in tasks])
                for task in tasks:
                    if task in self.task_list:
                        self.task_list.remove(task)
                self.task_list = (self.task_list[0:ind] + [heading] + tasks +
                                  self.task_list[ind:])
        return True

    def set_led_headings(self):
        if not self.should_transpose_view:
            new_headings = ['Name'] + self.point_strings
        else:
            new_headings = ['Point'] + self.task_list
        if new_headings == self.led_headings:
            return False
        self.led_headings = new_headings

        # Get a list of tasks belonging to families.
        if len(self.family_tree) > 0:
            all_sub_tasks = reduce(
                lambda x, y: x + y,
                [self.family_tree[key] for key in self.family_tree]
            )
        else:
            all_sub_tasks = []

        if self.should_transpose_view:
            self.led_treeview.set_headers_clickable(True)

        tvcs = self.led_treeview.get_columns()
        labels = []
        families = []
        for n in range(1, len(self.led_headings)):
            text = self.led_headings[n]
            tip = self.led_headings[n]
            if self.should_hide_headings:
                text = "..."
            label = gtk.Label(text)
            label.set_use_underline(False)
            label.set_angle(90)
            label.show()
            labels.append(label)
            label_box = gtk.VBox()

            # Functionality for collapsed families in transpose view.
            if self.should_transpose_view:
                self._set_heading_attributes(tvcs[n], label, text,
                                             all_sub_tasks)

            label_box.pack_start(label, expand=False, fill=False)
            label_box.show()
            self._set_tooltip(label_box, tip)
            tvcs[n].set_widget(label_box)

            if self.should_transpose_view and text in self.family_tree:
                families.append(n)

        max_pixel_length = -1
        for label in labels:
            x, y = label.get_layout().get_size()
            if x > max_pixel_length:
                max_pixel_length = x
        for label in labels:
            while label.get_layout().get_size()[0] < max_pixel_length:
                label.set_text(label.get_text() + ' ')

        # Pre-collapse all families.
        for family in families:
            self.transposed_label_click(None, family)

    def _set_heading_attributes(self, col, label, text, all_sub_tasks):
        """Sets the attributes required for folding columns."""
        # Clear previously set attributes.
        for attribute in ['gcylc_parent_to', 'gcylc_task', 'gcylc_folded',
                          'gcylc_heading_label', 'gcylc_represents_task']:
            if hasattr(col, attribute):
                delattr(col, attribute)
        col.set_visible(True)

        # Set new attributes.
        if self.should_group_families and (text in self.family_tree or
                                           text in all_sub_tasks):
            if text in self.family_tree:
                # Heading is a family.
                label.set_text(self.RIGHT_ARROW + " " + text)
                col.gcylc_parent_to = self.family_tree[text]
                col.gcylc_task = text
                col.gcylc_folded = text in self.expanded_rows
                col.gcylc_heading_label = label
            elif text in all_sub_tasks:
                # Heading is the child of a family.
                col.gcylc_represents_task = text
                label.set_text(' ' * self.INDENT + text)

    def transposed_label_click(self, widget, heading):
        """Click handler for column headings in transposed mode.
        Responsible for folding / unfolding family sections."""
        column = self.led_treeview.get_column(heading)
        if not hasattr(column, 'gcylc_folded'):
            # Column is not foldable.
            return False

        column.gcylc_folded = not column.gcylc_folded
        if hasattr(column, 'gcylc_parent_to'):
            # Column is for a family.
            for col in self.led_treeview.get_columns():
                if hasattr(col, 'gcylc_represents_task'):
                    if col.gcylc_represents_task in column.gcylc_parent_to:
                        col.set_visible(not column.gcylc_folded)

        # Toggle folding arrows.
        if column.gcylc_folded:
            if hasattr(column, 'gcylc_heading_label'):
                column.gcylc_heading_label.set_text(
                    self.RIGHT_ARROW +
                    column.gcylc_heading_label.get_text()[3:]
                )
        else:
            if hasattr(column, 'gcylc_heading_label'):
                column.gcylc_heading_label.set_text(
                    self.DOWN_ARROW +
                    column.gcylc_heading_label.get_text()[3:]
                )

    def _get_expanded_rows(self):
        """Updates list of currently expanded rows (or columns in transpose
        mode."""
        self.expanded_rows = []
        if self.is_transposed:
            for column in self.led_treeview.get_columns():
                if hasattr(column, 'gcylc_folded'):
                    if not column.gcylc_folded:
                        self.expanded_rows.append(column.gcylc_task)
        else:
            rows = []
            self.led_treeview.map_expanded_rows(
                lambda treeview, path, data: rows.append(path),
                None
            )
            for row in rows:
                self.expanded_rows.append(self.led_treestore.get_value(
                    self.led_treestore.get_iter(row), 0))

    def ledview_widgets(self):
        # Make note of expanded rows.
        self._get_expanded_rows()

        if not self.should_transpose_view:
            types = [str] + [gtk.gdk.Pixbuf] * len(self.point_strings)
            num_new_columns = len(types)
        else:
            types = [str] + [gtk.gdk.Pixbuf] * len(self.task_list) + [str]
            num_new_columns = 1 + len(self.task_list)
        new_led_treestore = gtk.TreeStore(*types)
        old_types = []
        for i in range(self.led_treestore.get_n_columns()):
            old_types.append(self.led_treestore.get_column_type(i))
        new_types = []
        for i in range(new_led_treestore.get_n_columns()):
            new_types.append(new_led_treestore.get_column_type(i))
        treeview_has_content = bool(len(self.led_treeview.get_columns()))

        if treeview_has_content and old_types == new_types:
            self.set_led_headings()
            self.led_treestore.clear()
            self.is_transposed = self.should_transpose_view
            return False

        self.led_treestore = new_led_treestore

        if (treeview_has_content and
                self.is_transposed == self.should_transpose_view):
            tvcs_for_removal = self.led_treeview.get_columns()[
                num_new_columns:]

            for tvc in tvcs_for_removal:
                self.led_treeview.remove_column(tvc)

            self.led_treeview.set_model(self.led_treestore)
            num_columns = len(self.led_treeview.get_columns())
            for model_col_num in range(num_columns, num_new_columns):
                # Add newly-needed columns.
                cr = gtk.CellRendererPixbuf()
                cr.set_property('xalign', 0)
                tvc = gtk.TreeViewColumn("")
                tvc.connect(
                    'clicked', self.transposed_label_click, model_col_num)
                tvc.pack_end(cr, True)
                tvc.set_attributes(cr, pixbuf=model_col_num)
                self.led_treeview.append_column(tvc)
            self.set_led_headings()
            return False

        tvcs = self.led_treeview.get_columns()
        for tvc in tvcs:
            self.led_treeview.remove_column(tvc)

        self.led_treeview.set_model(self.led_treestore)

        if not self.should_transpose_view:
            tvc = gtk.TreeViewColumn('Name')
        else:
            tvc = gtk.TreeViewColumn('Point')

        cr = gtk.CellRendererText()
        tvc.pack_start(cr, False)
        tvc.set_attributes(cr, text=0)

        self.led_treeview.append_column(tvc)

        if not self.should_transpose_view:
            data_range = range(1, len(self.point_strings) + 1)
        else:
            data_range = range(1, len(self.task_list) + 1)

        for n in data_range:
            cr = gtk.CellRendererPixbuf()
            cr.set_property('xalign', 0)
            tvc = gtk.TreeViewColumn("")
            tvc.pack_end(cr, True)
            tvc.set_attributes(cr, pixbuf=n)
            self.led_treeview.append_column(tvc)
            if self.should_transpose_view:
                tvc.connect('clicked', self.transposed_label_click, n)
        self.set_led_headings()
        self.is_transposed = self.should_transpose_view

    def on_query_tooltip(self, widget, x, y, kbd_ctx, tooltip):
        """Handle a tooltip creation request."""
        tip_context = self.led_treeview.get_tooltip_context(x, y, kbd_ctx)
        if tip_context is None:
            self._prev_tooltip_task_id = None
            return False
        x, y = self.led_treeview.convert_widget_to_bin_window_coords(x, y)
        path, column, cell_x, cell_y = self.led_treeview.get_path_at_pos(x, y)
        col_index = self.led_treeview.get_columns().index(column)
        if not self.is_transposed:
            iter_ = self.led_treeview.get_model().get_iter(path)
            name = self.led_treeview.get_model().get_value(iter_, 0)
            try:
                point_string = self.led_headings[col_index]
            except IndexError:
                # This can occur for a tooltip while switching from transposed.
                return False
            if col_index == 0:
                task_id = name
            else:
                task_id = TaskID.get(name, point_string)
        else:
            try:
                point_string = self.point_strings[path[0]]
            except IndexError:
                return False
            if col_index == 0:
                task_id = point_string
            else:
                try:
                    name = self.led_headings[col_index]
                except IndexError:
                    return False
                task_id = TaskID.get(name, point_string)
        if task_id != self._prev_tooltip_task_id:
            self._prev_tooltip_task_id = task_id
            tooltip.set_text(None)
            return False
        if col_index == 0:
            tooltip.set_text(task_id)
            return True
        text = get_id_summary(task_id, self.state_summary,
                              self.fam_state_summary, self.descendants)
        if text == task_id:
            return False
        tooltip.set_text(text)
        return True

    def update_gui(self):
        self.action_required = False
        state_summary = {}
        state_summary.update(self.state_summary)
        state_summary.update(self.fam_state_summary)
        self.ledview_widgets()

        tasks_by_point_string = {}
        tasks_by_name = {}
        for id_ in state_summary:
            name, point_string = TaskID.split(id_)
            tasks_by_point_string.setdefault(point_string, [])
            tasks_by_point_string[point_string].append(name)
            tasks_by_name.setdefault(name, [])
            tasks_by_name[name].append(point_string)

        names = tasks_by_name.keys()
        names.sort()

        if not self.is_transposed:
            self._update_gui_regular(tasks_by_name, state_summary)
        else:
            self._update_gui_transpose(tasks_by_point_string, state_summary)

        self.led_treeview.columns_autosize()
        return False

    def _update_gui_transpose(self, tasks_by_point_string, state_summary):
        """Logic for updating the gui in transpose mode."""
        for point_string in self.point_strings:
            tasks_at_point_string = tasks_by_point_string[point_string]
            state_list = []
            for name in self.task_list:
                task_id = TaskID.get(name, point_string)
                if task_id in self.fam_state_summary:
                    dot_type = 'family'
                else:
                    dot_type = 'task'
                if name in tasks_at_point_string:
                    state = state_summary[task_id]['state']
                    state_list.append(self.dots[dot_type][state])
                else:
                    state_list.append(self.dots[dot_type]['empty'])
            try:
                self.led_treestore.append(
                    None, row=[point_string] + state_list + [point_string])
            except ValueError:
                # A very laggy store can change the columns and raise this.
                return False

    def _update_gui_regular(self, tasks_by_name, state_summary):
        """Logic for updating the gui in regular mode."""
        children = []
        to_unfold = []
        parent_iter = None
        for name in self.task_list:
            point_strings_for_tasks = tasks_by_name.get(name, [])
            if not point_strings_for_tasks:
                continue
            state_list = []
            for point_string in self.point_strings:
                if point_string in point_strings_for_tasks:
                    task_id = TaskID.get(name, point_string)
                    state = state_summary[task_id]['state']
                    if task_id in self.fam_state_summary:
                        dot_type = 'family'
                    else:
                        dot_type = 'task'
                    state_list.append(self.dots[dot_type][state])
                else:
                    state_list.append(self.dots['task']['empty'])
            try:
                if name in self.family_tree:
                    # Task is a family.
                    self.led_treestore.append(
                        None, row=[name] + state_list)
                    children = self.family_tree[name]

                    # Get iter for this family's entry.
                    iter_ = self.led_treestore.get_iter_first()
                    temp = self.led_treestore.get_value(
                        iter_, 0)
                    while temp != name:
                        iter_ = self.led_treestore.iter_next(iter_)
                        temp = self.led_treestore.get_value(iter_, 0)
                    parent_iter = iter_

                    # Unfold if family was folded before update
                    if name in self.expanded_rows:
                        to_unfold.append(
                            self.led_treestore.get_path(iter_))

                elif name in children:
                    # Task belongs to a family.
                    self.led_treestore.append(
                        parent_iter, row=[name] + state_list)

                else:
                    # Task does not belong to a family.
                    self.led_treestore.append(
                        None, row=[name] + state_list)
            except ValueError:
                # A very laggy store can change the columns and raise this.
                return False

        # Unfold any rows that were unfolded before the update.
        for path in to_unfold:
            self.led_treeview.expand_row(path, True)

    def run(self):
        while not self.quit:
            if self.update() or self.action_required:
                gobject.idle_add(self.update_gui)
            sleep(0.2)
        else:
            pass
